import abc
from dataclasses import dataclass
from enum import Enum, Flag, IntEnum, IntFlag
from sys import intern
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Optional,
    Protocol,
    Type,
    TypeAlias,
    TypeVar,
    Union,
    cast,
)

from .drawing_primitives import convert_to_device_color
from .syntax import Name, wrap_in_local_context

if TYPE_CHECKING:
    from .drawing_primitives import Color, DeviceCMYK, DeviceGray, DeviceRGB
    from .fpdf import FPDF


class SignatureFlag(IntEnum):
    SIGNATURES_EXIST = 1
    "If set, the document contains at least one signature field."
    APPEND_ONLY = 2
    """
    If set, the document contains signatures that may be invalidated
    if the file is saved (written) in a way that alters its previous contents,
    as opposed to an incremental update.
    """


E = TypeVar("E", bound="CoerciveEnum")
IE = TypeVar("IE", bound="CoerciveIntEnum")
IF = TypeVar("IF", bound="CoerciveIntFlag")


class CoerciveEnum(Enum):
    "An enumeration that provides a helper to coerce strings into enumeration members."

    @classmethod
    def coerce(cls: Type[E], value: E | str, case_sensitive: bool = False) -> E:
        """
        Attempt to coerce `value` into a member of this enumeration.

        If value is already a member of this enumeration it is returned unchanged.
        Otherwise, if it is a string, attempt to convert it as an enumeration value. If
        that fails, attempt to convert it (case insensitively, by upcasing) as an
        enumeration name.

        If all different conversion attempts fail, an exception is raised.

        Args:
            value (Enum, str): the value to be coerced.

        Raises:
            ValueError: if `value` is a string but neither a member by name nor value.
            TypeError: if `value`'s type is neither a member of the enumeration nor a
                string.
        """

        if isinstance(value, cls):
            return value

        if isinstance(value, str):
            try:
                return cls(value)
            except ValueError:
                pass
            try:
                return cls[value] if case_sensitive else cls[value.upper()]
            except KeyError:
                pass

            raise ValueError(f"{value} is not a valid {cls.__name__}")

        raise TypeError(f"{value} cannot be converted to a {cls.__name__}")


class CoerciveIntEnum(IntEnum):
    """
    An enumeration that provides a helper to coerce strings and integers into
    enumeration members.
    """

    @classmethod
    def coerce(cls: Type[IE], value: IE | str | int) -> IE:
        """
        Attempt to coerce `value` into a member of this enumeration.

        If value is already a member of this enumeration it is returned unchanged.
        Otherwise, if it is a string, attempt to convert it (case insensitively, by
        upcasing) as an enumeration name. Otherwise, if it is an int, attempt to
        convert it as an enumeration value.

        Otherwise, an exception is raised.

        Args:
            value (IntEnum, str, int): the value to be coerced.

        Raises:
            ValueError: if `value` is an int but not a member of this enumeration.
            ValueError: if `value` is a string but not a member by name.
            TypeError: if `value`'s type is neither a member of the enumeration nor an
                int or a string.
        """
        if isinstance(value, cls):
            return value

        if isinstance(value, str):
            try:
                return cls[value.upper()]
            except KeyError:
                raise ValueError(f"{value} is not a valid {cls.__name__}") from None

        if isinstance(value, int):
            return cls(value)

        raise TypeError(f"{value} cannot convert to a {cls.__name__}")


class CoerciveIntFlag(IntFlag):
    """
    Enumerated constants that can be combined using the bitwise operators,
    with a helper to coerce strings and integers into enumeration members.
    """

    @classmethod
    def coerce(cls: Type[IF], value: IF | str | int) -> IF:
        """
        Attempt to coerce `value` into a member of this enumeration.

        If value is already a member of this enumeration it is returned unchanged.
        Otherwise, if it is a string, attempt to convert it (case insensitively, by
        upcasing) as an enumeration name. Otherwise, if it is an int, attempt to
        convert it as an enumeration value.
        Otherwise, an exception is raised.

        Args:
            value (IntEnum, str, int): the value to be coerced.

        Raises:
            ValueError: if `value` is an int but not a member of this enumeration.
            ValueError: if `value` is a string but not a member by name.
            TypeError: if `value`'s type is neither a member of the enumeration nor an
                int or a string.
        """
        if isinstance(value, cls):
            return value

        if isinstance(value, str):
            try:
                return cls[value.upper()]
            except KeyError:
                pass
            try:
                flags = cls[value[0].upper()]
                for char in value[1:]:
                    flags = flags | cls[char.upper()]
                return flags
            except KeyError:
                raise ValueError(f"{value} is not a valid {cls.__name__}") from None

        if isinstance(value, int):
            return cls(value)

        raise TypeError(f"{value} cannot convert to a {cls.__name__}")


class WrapMode(CoerciveEnum):
    "Defines how to break and wrap lines in multi-line text."

    WORD = intern("WORD")
    "Wrap by word"

    CHAR = intern("CHAR")
    "Wrap by character"


class CharVPos(CoerciveEnum):
    "Defines the vertical position of text relative to the line."

    SUP = intern("SUP")
    "Superscript"

    SUB = intern("SUB")
    "Subscript"

    NOM = intern("NOM")
    "Nominator of a fraction"

    DENOM = intern("DENOM")
    "Denominator of a fraction"

    LINE = intern("LINE")
    "Default line position"


class Align(CoerciveEnum):
    "Defines how to render text in a cell"

    C = intern("CENTER")
    "Center text horizontally"

    X = intern("X_CENTER")
    "Center text horizontally around current x position"

    L = intern("LEFT")
    "Left-align text"

    R = intern("RIGHT")
    "Right-align text"

    J = intern("JUSTIFY")
    "Justify text"

    @classmethod
    def coerce(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls, value: Union["Align", str], case_sensitive: bool = False
    ) -> "Align":
        if value == "":
            return cls.L
        if isinstance(value, str):
            value = value.upper()
        return super(cls, cls).coerce(
            value, case_sensitive  # pyright: ignore[reportArgumentType]
        )


class VAlign(CoerciveEnum):
    """Defines how to vertically render text in a cell.
    Default value is MIDDLE"""

    M = intern("MIDDLE")
    "Center text vertically"

    T = intern("TOP")
    "Place text at the top of the cell, but obey the cells padding"

    B = intern("BOTTOM")
    "Place text at the bottom of the cell, but obey the cells padding"

    @classmethod
    def coerce(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls, value: Union["VAlign", str], case_sensitive: bool = False
    ) -> "VAlign":
        if value == "":
            return cls.M
        return super(cls, cls).coerce(
            value, case_sensitive  # pyright: ignore[reportArgumentType]
        )


class TextEmphasis(CoerciveIntFlag):
    """
    Indicates use of bold / italics / underline.

    This enum values can be combined with & and | operators:
        style = B | I
    """

    NONE = 0
    "No emphasis"

    B = 1
    "Bold"

    I = 2
    "Italics"

    U = 4
    "Underline"

    S = 8
    "Strikethrough"

    @property
    def style(self) -> str:
        return "".join(
            name for name, value in self.__class__.__members__.items() if value & self
        )

    def add(self, value: "TextEmphasis") -> "TextEmphasis":
        return self | value

    def remove(self, value: "TextEmphasis") -> "TextEmphasis":
        return TextEmphasis.coerce(
            "".join(s for s in self.style if s not in value.style)
        )

    @classmethod
    def coerce(cls, value: Union["TextEmphasis", str, int]) -> "TextEmphasis":
        if isinstance(value, str):
            if value == "":
                return cls.NONE
            if value.upper() == "BOLD":
                return cls.B
            if value.upper() == "ITALICS":
                return cls.I
            if value.upper() == "UNDERLINE":
                return cls.U
            if value.upper() == "STRIKETHROUGH":
                return cls.S
        return super(cls, cls).coerce(value)


class MethodReturnValue(CoerciveIntFlag):
    """
    Defines the return value(s) of a FPDF content-rendering method.

    This enum values can be combined with & and | operators:
        PAGE_BREAK | LINES
    """

    PAGE_BREAK = 1
    "The method will return a boolean indicating if a page break occurred"

    LINES = 2
    "The method will return a multi-lines array of strings, after performing word-wrapping"

    HEIGHT = 4
    "The method will return how much vertical space was used"


class CellBordersLayout(CoerciveIntFlag):
    """Defines how to render cell borders in table

    The integer value of `border` determines which borders are applied. Below are some common examples:

    - border=1 (LEFT): Only the left border is enabled.
    - border=3 (LEFT | RIGHT): Both the left and right borders are enabled.
    - border=5 (LEFT | TOP): The left and top borders are enabled.
    - border=12 (TOP | BOTTOM): The top and bottom borders are enabled.
    - border=15 (ALL): All borders (left, right, top, bottom) are enabled.
    - border=16 (INHERIT): Inherit the border settings from the parent element.

    Using `border=3` will combine LEFT and RIGHT borders, as it represents the
    bitwise OR of `LEFT (1)` and `RIGHT (2)`.
    """

    NONE = 0
    "Draw no border on any side of cell"

    LEFT = 1
    "Draw border on the left side of the cell"

    RIGHT = 2
    "Draw border on the right side of the cell"

    TOP = 4
    "Draw border on the top side of the cell"

    BOTTOM = 8
    "Draw border on the bottom side of the cell"

    ALL = LEFT | RIGHT | TOP | BOTTOM
    "Draw border on all side of the cell"

    INHERIT = 16
    "Inherits the border layout from the table borders layout"

    @classmethod
    def coerce(cls, value: Union["CellBordersLayout", str, int]) -> "CellBordersLayout":
        if isinstance(value, int) and value > 16:
            raise ValueError("INHERIT cannot be combined with other values")
        return super().coerce(value)

    def __and__(self, value: int) -> "CellBordersLayout":
        value = super().__and__(value)
        if value > 16:
            raise ValueError("INHERIT cannot be combined with other values")
        return value

    def __or__(self, value: int) -> "CellBordersLayout":
        value = super().__or__(value)
        if value > 16:
            raise ValueError("INHERIT cannot be combined with other values")
        return value

    def __str__(self) -> str:
        border_str: list[str] = []
        if self & CellBordersLayout.LEFT:
            border_str.append("L")
        if self & CellBordersLayout.RIGHT:
            border_str.append("R")
        if self & CellBordersLayout.TOP:
            border_str.append("T")
        if self & CellBordersLayout.BOTTOM:
            border_str.append("B")
        return "".join(border_str) if border_str else "NONE"


@dataclass
class TableBorderStyle:
    """A helper class for drawing one border of a table

    Attributes:
        thickness: The thickness of the border. If None use default. If <= 0 don't draw the border.
        color: The color of the border. If None use default.
    """

    thickness: Optional[float] = None
    color: Union[
        int, tuple[int, int, int], "DeviceRGB", "DeviceGray", "DeviceCMYK", None
    ] = None
    dash: Optional[float] = None
    gap: float = 0.0
    phase: float = 0.0

    @staticmethod
    def from_bool(should_draw: Union[bool, "TableBorderStyle"]) -> "TableBorderStyle":
        """
        From boolean or TableBorderStyle input, convert to definite TableBorderStyle class object
        """
        if isinstance(should_draw, TableBorderStyle):
            return should_draw  # don't change specified TableBorderStyle
        if should_draw:
            return TableBorderStyle()  # keep default stroke
        return TableBorderStyle(thickness=0.0)  # don't draw the border

    def _changes_thickness(self, pdf: "FPDF") -> bool:
        """Return True if this style changes the thickness of the draw command, False otherwise"""
        return (
            self.thickness is not None
            and self.thickness > 0.0
            and self.thickness != pdf.line_width
        )

    def _changes_color(self, pdf: "FPDF") -> bool:
        """Return True if this style changes the color of the draw command, False otherwise"""
        return self.color is not None and self.color != pdf.draw_color

    @property
    def dash_dict(self) -> dict[str, Optional[float]]:
        """Return dict object specifying dash in the same format as the pdf object"""
        return {"dash": self.dash, "gap": self.gap, "phase": self.phase}

    def _changes_dash(self, pdf: "FPDF") -> bool:
        """Return True if this style changes the dash of the draw command, False otherwise"""
        return self.dash is not None and self.dash_dict != pdf.dash_pattern

    def changes_stroke(self, pdf: "FPDF") -> bool:
        """Return True if this style changes the any aspect of the draw command, False otherwise"""
        return self.should_render() and (
            self._changes_color(pdf)
            or self._changes_thickness(pdf)
            or self._changes_dash(pdf)
        )

    def should_render(self) -> bool:
        """Return True if this style produces a visible stroke, False otherwise"""
        return self.thickness is None or self.thickness > 0.0

    def _get_change_thickness_command(
        self, scale: float, pdf: Optional["FPDF"] = None
    ) -> list[str]:
        """Return list with string for the draw command to change thickness (empty if no change)"""
        thickness = self.thickness if pdf is None else pdf.line_width
        return [] if thickness is None else [f"{thickness * scale:.2f} w"]

    def _get_change_line_color_command(self, pdf: Optional["FPDF"] = None) -> list[str]:
        """Return list with string for the draw command to change color (empty if no change)"""
        if pdf is None:
            color = self.color
        else:
            color = pdf.draw_color
        return (
            []
            if color is None
            else [convert_to_device_color(color).serialize().upper()]
        )

    def _get_change_dash_command(
        self, scale: float, pdf: Optional["FPDF"] = None
    ) -> list[str]:
        """Return list with string for the draw command to change dash (empty if no change)"""
        dash_dict = self.dash_dict if pdf is None else pdf.dash_pattern
        dash, gap, phase = dash_dict["dash"], dash_dict["gap"], dash_dict["phase"]
        if dash is None:
            return []
        if dash <= 0:
            return ["[] 0 d"]
        assert phase is not None
        if gap is None or gap <= 0:
            return [f"[{dash * scale:.3f}] {phase * scale:.3f} d"]
        return [f"[{dash * scale:.3f} {gap * scale:.3f}] {phase * scale:.3f} d"]

    def get_change_stroke_commands(self, scale: float) -> list[str]:
        """Return list of strings for the draw command to change stroke (empty if no change)"""
        return (
            self._get_change_dash_command(scale)
            + self._get_change_line_color_command()
            + self._get_change_thickness_command(scale)
        )

    @staticmethod
    def get_line_command(x1: float, y1: float, x2: float, y2: float) -> list[str]:
        """Return list with string for the command to draw a line at the specified endpoints"""
        return [f"{x1:.2f} {y1:.2f} m {x2:.2f} {y2:.2f} l S"]

    def get_draw_commands(
        self, pdf: "FPDF", x1: float, y1: float, x2: float, y2: float
    ) -> list[str]:
        """
        Get draw commands for this section of a cell border. x and y are presumed to be already
        shifted and scaled.
        """
        if not self.should_render():
            return []

        if self.changes_stroke(pdf):
            draw_commands = self.get_change_stroke_commands(
                scale=pdf.k
            ) + self.get_line_command(x1, y1, x2, y2)
            # wrap in local context to prevent stroke changes from affecting later rendering
            return wrap_in_local_context(draw_commands)
        return self.get_line_command(x1, y1, x2, y2)


@dataclass
class TableCellStyle:
    """A helper class for drawing all the borders of one cell in a table

    Attributes:
        left: bool or TableBorderStyle specifying the style of the cell's left border
        bottom: bool or TableBorderStyle specifying the style of the cell's bottom border
        right: bool or TableBorderStyle specifying the style of the cell's right border
        top: bool or TableBorderStyle specifying the style of the cell's top border
    """

    left: bool | TableBorderStyle = False
    bottom: bool | TableBorderStyle = False
    right: bool | TableBorderStyle = False
    top: bool | TableBorderStyle = False

    def _get_common_border_style(self) -> Optional[bool | TableBorderStyle]:
        """Return bool or TableBorderStyle if all borders have the same style, otherwise None"""
        if all(
            isinstance(border, bool)
            for border in [self.left, self.bottom, self.right, self.top]
        ):
            if all(border for border in [self.left, self.bottom, self.right, self.top]):
                return True
            if all(
                not border for border in [self.left, self.bottom, self.right, self.top]
            ):
                return False
        elif all(
            isinstance(border, TableBorderStyle)
            for border in [self.left, self.bottom, self.right, self.top]
        ):
            common = self.left
            if all(border == common for border in [self.bottom, self.right, self.top]):
                return common
        return None

    @staticmethod
    def get_change_fill_color_command(color: Union["Color", str]) -> list[str]:
        """Return list with string for command to change device color (empty list if no color)"""
        return (
            []
            if color is None
            else [convert_to_device_color(color).serialize().lower()]
        )

    def get_draw_commands(
        self,
        pdf: "FPDF",
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        fill_color: Optional[Union["Color", str]] = None,
    ) -> list[str]:
        """
        Get list of primitive commands to draw the cell border for this cell, and fill it with the
        given fill color.
        """
        # y top to bottom instead of bottom to top
        y1 = pdf.h - y1
        y2 = pdf.h - y2
        # scale coordinates and thickness
        scale = pdf.k
        x1 *= scale
        y1 *= scale
        x2 *= scale
        y2 *= scale

        common_border_style = self._get_common_border_style()
        draw_commands, needs_wrap = (
            self._draw_when_no_common_style(x1, y1, x2, y2, pdf, fill_color)
            if common_border_style is None
            else (
                self._draw_with_no_border(x1, y1, x2, y2, pdf, fill_color)
                if common_border_style is False
                else self._draw_all_borders_the_same(
                    x1, y1, x2, y2, pdf, fill_color, scale, common_border_style
                )
            )
        )

        if needs_wrap:
            draw_commands = wrap_in_local_context(draw_commands)

        return draw_commands

    def _draw_when_no_common_style(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        pdf: "FPDF",
        fill_color: Optional[Union["Color", str]],
    ) -> tuple[list[str], bool]:
        """Get draw commands for case when some of the borders have different styles"""
        needs_wrap = False
        draw_commands: list[str] = []
        if fill_color is not None:
            # draw fill with no box
            if fill_color != pdf.fill_color:
                needs_wrap = True
                draw_commands.extend(self.get_change_fill_color_command(fill_color))
            draw_commands.append(f"{x1:.2f} {y2:.2f} {x2 - x1:.2f} {y1 - y2:.2f} re f")
        # draw the individual borders
        draw_commands.extend(
            TableBorderStyle.from_bool(self.left).get_draw_commands(pdf, x1, y2, x1, y1)
            + TableBorderStyle.from_bool(self.bottom).get_draw_commands(
                pdf, x1, y2, x2, y2
            )
            + TableBorderStyle.from_bool(self.right).get_draw_commands(
                pdf, x2, y2, x2, y1
            )
            + TableBorderStyle.from_bool(self.top).get_draw_commands(
                pdf, x1, y1, x2, y1
            )
        )
        return draw_commands, needs_wrap

    def _draw_with_no_border(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        pdf: "FPDF",
        fill_color: Optional[Union["Color", str]],
    ) -> tuple[list[str], bool]:
        """Get draw commands for case when all of the borders are off / not drawn"""
        needs_wrap = False
        draw_commands: list[str] = []
        if fill_color is not None:
            # draw fill with no box
            if fill_color != pdf.fill_color:
                needs_wrap = True
                draw_commands.extend(self.get_change_fill_color_command(fill_color))
            draw_commands.append(f"{x1:.2f} {y2:.2f} {x2 - x1:.2f} {y1 - y2:.2f} re f")
        return draw_commands, needs_wrap

    def _draw_all_borders_the_same(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        pdf: "FPDF",
        fill_color: Optional[Union["Color", str]],
        scale: float,
        common_border_style: Optional[TableBorderStyle | bool],
    ) -> tuple[list[str], bool]:
        """Get draw commands for case when all the borders have the same style"""
        needs_wrap = False
        draw_commands: list[str] = []
        # all borders are the same
        if isinstance(
            common_border_style, TableBorderStyle
        ) and common_border_style.changes_stroke(pdf):
            # the border styles aren't default, so
            draw_commands.extend(common_border_style.get_change_stroke_commands(scale))
            needs_wrap = True
        if fill_color is not None:
            # draw filled rectangle
            if fill_color != pdf.fill_color:
                needs_wrap = True
                draw_commands.extend(self.get_change_fill_color_command(fill_color))
            draw_commands.append(f"{x1:.2f} {y2:.2f} {x2 - x1:.2f} {y1 - y2:.2f} re B")
        else:
            # draw empty rectangle
            draw_commands.append(f"{x1:.2f} {y2:.2f} {x2 - x1:.2f} {y1 - y2:.2f} re S")
        return draw_commands, needs_wrap

    def override_cell_border(self, cell_border: CellBordersLayout) -> "TableCellStyle":
        """Allow override by CellBordersLayout mechanism"""
        return (
            self
            if cell_border == CellBordersLayout.INHERIT
            else TableCellStyle(  # translate cell_border into equivalent TableCellStyle
                left=bool(cell_border & CellBordersLayout.LEFT),
                bottom=bool(cell_border & CellBordersLayout.BOTTOM),
                right=bool(cell_border & CellBordersLayout.RIGHT),
                top=bool(cell_border & CellBordersLayout.TOP),
            )
        )

    def draw_cell_border(
        self,
        pdf: "FPDF",
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        fill_color: Optional[Union["Color", str]] = None,
    ) -> None:
        """
        Draw the cell border for this cell, and fill it with the given fill color.
        """
        pdf._out(  # pylint: disable=protected-access # pyright: ignore[reportPrivateUsage]
            " ".join(self.get_draw_commands(pdf, x1, y1, x2, y2, fill_color=fill_color))
        )


class TableBordersLayout(abc.ABC):
    """
    Customizable class for setting the drawing style of cell borders for the whole table.
    cell_style_getter is an abstract method that derived classes must implement. All current classes
    do not use self, but it is available in case a very complicated derived class needs to refer to
    stored internal data.

    Standard TableBordersLayouts are available as static members of this class

    Attributes:
        cell_style_getter: a callable that takes row_num, column_num,
            num_heading_rows, num_rows, num_columns; and returns the drawing style of
            the cell border (as a TableCellStyle object)
        ALL: static TableBordersLayout that draws all table cells borders
        NONE: static TableBordersLayout that draws no table cells borders
        INTERNAL: static TableBordersLayout that draws only internal horizontal & vertical borders
        MINIMAL: static TableBordersLayout that draws only the top horizontal border, below the
            headings, and internal vertical borders
        HORIZONTAL_LINES: static TableBordersLayout that draws only horizontal lines
        NO_HORIZONTAL_LINES: static TableBordersLayout that draws all cells border except interior
            horizontal lines after the headings
        SINGLE_TOP_LINE: static TableBordersLayout that draws only the top horizontal border, below
            the headings
    """

    ALL: ClassVar["TableBordersLayout"]
    NONE: ClassVar["TableBordersLayout"]
    INTERNAL: ClassVar["TableBordersLayout"]
    MINIMAL: ClassVar["TableBordersLayout"]
    HORIZONTAL_LINES: ClassVar["TableBordersLayout"]
    NO_HORIZONTAL_LINES: ClassVar["TableBordersLayout"]
    SINGLE_TOP_LINE: ClassVar["TableBordersLayout"]

    @abc.abstractmethod
    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        """Specify the desired TableCellStyle for the given position in the table

        Args:
            row_idx: the 0-based index of the row in the table
            col_idx: the 0-based logical index of the cell in the row. If colspan > 1, this indexes
                into non-null cells. e.g. if there are two cells with colspan = 3, then col_idx will
                be 0 or 1
            col_pos: the 0-based physical position of the cell in the row. If colspan > 1, this
                indexes into all cells including null ones. e.g. e.g. if there are two cells with
                colspan = 3, then col_pos will be 0 or 3
            num_heading_rows: the number of rows in the table heading
            num_rows: the total number of rows in the table
            num_col_idx: the number of non-null cells. e.g. if there are two cells with colspan = 3,
                then num_col_idx = 2
            num_col_pos: the full width of the table in physical cells. e.g. if there are two cells
                with colspan = 3, then num_col_pos = 6
        Returns:
            TableCellStyle for the given position in the table
        """
        raise NotImplementedError

    @classmethod
    def coerce(cls, value: Union["TableBordersLayout", str]) -> "TableBordersLayout":
        """
        Attempt to coerce `value` into a member of this class.

        If value is already a member of this enumeration it is returned unchanged.
        Otherwise, if it is a string, attempt to convert it as an enumeration value. If
        that fails, attempt to convert it (case insensitively, by upcasing) as an
        enumeration name.

        If all different conversion attempts fail, an exception is raised.

        Args:
            value (Enum, str): the value to be coerced.

        Raises:
            ValueError: if `value` is a string but neither a member by name nor value.
            TypeError: if `value`'s type is neither a member of the enumeration nor a
                string.
        """

        if isinstance(value, cls):
            return value

        if isinstance(value, str):
            try:
                coerced_value = getattr(cls, value.upper())
                if isinstance(coerced_value, cls):
                    return coerced_value
            except ValueError:
                pass

        raise ValueError(f"{value} is not a valid {cls.__name__}")


class TableBordersLayoutAll(TableBordersLayout):
    """Class for drawing all cell borders"""

    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        return TableCellStyle(left=True, bottom=True, right=True, top=True)


# add as static member of base TableBordersLayout class
TableBordersLayout.ALL = TableBordersLayoutAll()


class TableBordersLayoutNone(TableBordersLayout):
    """Class for drawing zero cell borders"""

    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        return TableCellStyle(left=False, bottom=False, right=False, top=False)


# add as static member of base TableBordersLayout class
TableBordersLayout.NONE = TableBordersLayoutNone()


class TableBordersLayoutInternal(TableBordersLayout):
    """Class to draw only internal horizontal & vertical borders"""

    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        return TableCellStyle(
            left=col_idx > 0,
            bottom=row_idx < num_rows - 1,
            right=col_idx < num_col_idx - 1,
            top=row_idx > 0,
        )


# add as static member of base TableBordersLayout class
TableBordersLayout.INTERNAL = TableBordersLayoutInternal()


class TableBordersLayoutMinimal(TableBordersLayout):
    """
    Class to draw only the top horizontal border, below the headings, and internal vertical borders
    """

    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        return TableCellStyle(
            left=col_idx > 0,
            bottom=row_idx < num_heading_rows,
            right=col_idx < num_col_idx - 1,
            top=0 < row_idx <= num_heading_rows,
        )


# add as static member of base TableBordersLayout class
TableBordersLayout.MINIMAL = TableBordersLayoutMinimal()


class TableBordersLayoutHorizontalLines(TableBordersLayout):
    """Class to draw only horizontal lines"""

    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        return TableCellStyle(
            left=False,
            bottom=row_idx < num_rows - 1,
            right=False,
            top=row_idx > 0,
        )


# add as static member of base TableBordersLayout class
TableBordersLayout.HORIZONTAL_LINES = TableBordersLayoutHorizontalLines()


class TableBordersLayoutNoHorizontalLines(TableBordersLayout):
    """Class to draw all cells border except interior horizontal lines after the headings"""

    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        return TableCellStyle(
            left=True,
            bottom=row_idx == num_rows - 1,
            right=True,
            top=row_idx <= num_heading_rows,
        )


# add as static member of base TableBordersLayout class
TableBordersLayout.NO_HORIZONTAL_LINES = TableBordersLayoutNoHorizontalLines()


class TableBordersLayoutSingleTopLine(TableBordersLayout):
    """Class to draw a single top line"""

    def cell_style_getter(
        self,
        row_idx: int,
        col_idx: int,
        col_pos: int,
        num_heading_rows: int,
        num_rows: int,
        num_col_idx: int,
        num_col_pos: int,
    ) -> TableCellStyle:
        return TableCellStyle(
            left=False, bottom=row_idx <= num_heading_rows - 1, right=False, top=False
        )


# add as static member of base TableBordersLayout class
TableBordersLayout.SINGLE_TOP_LINE = TableBordersLayoutSingleTopLine()


class CellFillProtocol(Protocol):
    """Protocol for custom table cell fill mode classes"""

    def should_fill_cell(self, i: int, j: int) -> bool: ...


TableCellFillModeType: TypeAlias = "TableCellFillMode | CellFillProtocol"


class TableCellFillMode(CoerciveEnum):
    "Defines which table cells to fill"

    NONE = intern("NONE")
    "Fill zero table cell"

    ALL = intern("ALL")
    "Fill all table cells"

    ROWS = intern("ROWS")
    "Fill only table cells in odd rows"

    COLUMNS = intern("COLUMNS")
    "Fill only table cells in odd columns"

    EVEN_ROWS = intern("EVEN_ROWS")
    "Fill only table cells in even rows"

    EVEN_COLUMNS = intern("EVEN_COLUMNS")
    "Fill only table cells in even columns"

    @classmethod
    def coerce(  # type: ignore[override]
        cls,
        value: Union[TableCellFillModeType, str],
        case_sensitive: bool = False,
    ) -> TableCellFillModeType:
        if callable(getattr(value, "should_fill_cell", None)):
            return cast(CellFillProtocol, value)
        return super().coerce(value, case_sensitive)  # type: ignore[arg-type] # pyright: ignore[reportArgumentType]

    def should_fill_cell(self, i: int, j: int) -> bool:
        if self is TableCellFillMode.NONE:
            return False
        if self is TableCellFillMode.ALL:
            return True
        if self is TableCellFillMode.ROWS:
            return i % 2 == 1
        if self is TableCellFillMode.COLUMNS:
            return j % 2 == 1
        if self is TableCellFillMode.EVEN_ROWS:
            return i % 2 == 0
        if self is TableCellFillMode.EVEN_COLUMNS:
            return j % 2 == 0
        raise NotImplementedError


class TableSpan(CoerciveEnum):
    ROW = intern("ROW")
    "Mark this cell as a continuation of the previous row"

    COL = intern("COL")
    "Mark this cell as a continuation of the previous column"


class TableHeadingsDisplay(CoerciveIntEnum):
    "Defines how the table headings should be displayed"

    NONE = 0
    "0: Only render the table headings at the beginning of the table"

    ON_TOP_OF_EVERY_PAGE = 1
    "1: When a page break occurs, repeat the table headings at the top of every table fragment"


class RenderStyle(CoerciveEnum):
    "Defines how to render shapes"

    D = intern("DRAW")
    """
    Draw lines.
    Line color can be controlled with `fpdf.fpdf.FPDF.set_draw_color()`.
    Line thickness can be controlled with `fpdf.fpdf.FPDF.set_line_width()`.
    """

    F = intern("FILL")
    """
    Fill areas.
    Filling color can be controlled with `fpdf.fpdf.FPDF.set_fill_color()`.
    """

    DF = intern("DRAW_FILL")
    "Draw lines and fill areas"

    @property
    def operator(self) -> str:
        return {RenderStyle.D: "S", RenderStyle.F: "f", RenderStyle.DF: "B"}[self]

    @property
    def is_draw(self) -> bool:
        return self in (RenderStyle.D, RenderStyle.DF)

    @property
    def is_fill(self) -> bool:
        return self in (RenderStyle.F, RenderStyle.DF)

    @classmethod
    def coerce(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls, value: Union["RenderStyle", str], case_sensitive: bool = False
    ) -> "RenderStyle":
        if not value:
            return cls.D
        if value == "FD":
            value = "DF"
        return super(cls, cls).coerce(
            value, case_sensitive  # pyright: ignore[reportArgumentType]
        )


class TextMode(CoerciveIntEnum):
    "Values described in PDF spec section 'Text Rendering Mode'"

    FILL = 0
    STROKE = 1
    FILL_STROKE = 2
    INVISIBLE = 3
    FILL_CLIP = 4
    STROKE_CLIP = 5
    FILL_STROKE_CLIP = 6
    CLIP = 7


class XPos(CoerciveEnum):
    "Positional values in horizontal direction for use after printing text."

    LEFT = intern("LEFT")  # self.x
    "left end of the cell"

    RIGHT = intern("RIGHT")  # self.x + w
    "right end of the cell (default)"

    START = intern("START")
    "left start of actual text"

    END = intern("END")
    "right end of actual text"

    WCONT = intern("WCONT")
    "for write() to continue next (slightly left of END)"

    CENTER = intern("CENTER")
    "center of actual text"

    LMARGIN = intern("LMARGIN")  # self.l_margin
    "left page margin (start of printable area)"

    RMARGIN = intern("RMARGIN")  # self.w - self.r_margin
    "right page margin (end of printable area)"


class YPos(CoerciveEnum):
    "Positional values in vertical direction for use after printing text"

    TOP = intern("TOP")  # self.y
    "top of the first line (default)"

    LAST = intern("LAST")
    "top of the last line (same as TOP for single-line text)"

    NEXT = intern("NEXT")  # LAST + h
    "top of next line (bottom of current text)"

    TMARGIN = intern("TMARGIN")  # self.t_margin
    "top page margin (start of printable area)"

    BMARGIN = intern("BMARGIN")  # self.h - self.b_margin
    "bottom page margin (end of printable area)"


class Angle(CoerciveIntEnum):
    "Direction values used for mirror transformations specifying the angle of mirror line"

    NORTH = 90
    EAST = 0
    SOUTH = 270
    WEST = 180
    NORTHEAST = 45
    SOUTHEAST = 315
    SOUTHWEST = 225
    NORTHWEST = 135


class PageLayout(CoerciveEnum):
    "Specify the page layout shall be used when the document is opened"

    SINGLE_PAGE = Name("SinglePage")
    "Display one page at a time"

    ONE_COLUMN = Name("OneColumn")
    "Display the pages in one column"

    TWO_COLUMN_LEFT = Name("TwoColumnLeft")
    "Display the pages in two columns, with odd-numbered pages on the left"

    TWO_COLUMN_RIGHT = Name("TwoColumnRight")
    "Display the pages in two columns, with odd-numbered pages on the right"

    TWO_PAGE_LEFT = Name("TwoPageLeft")
    "Display the pages two at a time, with odd-numbered pages on the left"

    TWO_PAGE_RIGHT = Name("TwoPageRight")
    "Display the pages two at a time, with odd-numbered pages on the right"


class PageMode(CoerciveEnum):
    "Specifying how to display the document on exiting full-screen mode"

    USE_NONE = Name("UseNone")
    "Neither document outline nor thumbnail images visible"

    USE_OUTLINES = Name("UseOutlines")
    "Document outline visible"

    USE_THUMBS = Name("UseThumbs")
    "Thumbnail images visible"

    FULL_SCREEN = Name("FullScreen")
    "Full-screen mode, with no menu bar, window controls, or any other window visible"

    USE_OC = Name("UseOC")
    "Optional content group panel visible"

    USE_ATTACHMENTS = Name("UseAttachments")
    "Attachments panel visible"


class TextMarkupType(CoerciveEnum):
    "Subtype of a text markup annotation"

    HIGHLIGHT = Name("Highlight")

    UNDERLINE = Name("Underline")

    SQUIGGLY = Name("Squiggly")

    STRIKE_OUT = Name("StrikeOut")


class BlendMode(CoerciveEnum):
    "An enumeration of the named standard named blend functions supported by PDF."

    NORMAL = Name("Normal")
    '''"Selects the source color, ignoring the backdrop."'''
    MULTIPLY = Name("Multiply")
    '''"Multiplies the backdrop and source color values."'''
    SCREEN = Name("Screen")
    """
    "Multiplies the complements of the backdrop and source color values, then
    complements the result."
    """
    OVERLAY = Name("Overlay")
    """
    "Multiplies or screens the colors, depending on the backdrop color value. Source
    colors overlay the backdrop while preserving its highlights and shadows. The
    backdrop color is not replaced but is mixed with the source color to reflect the
    lightness or darkness of the backdrop."
    """
    DARKEN = Name("Darken")
    '''"Selects the darker of the backdrop and source colors."'''
    LIGHTEN = Name("Lighten")
    '''"Selects the lighter of the backdrop and source colors."'''
    COLOR_DODGE = Name("ColorDodge")
    """
    "Brightens the backdrop color to reflect the source color. Painting with black
     produces no changes."
    """
    COLOR_BURN = Name("ColorBurn")
    """
    "Darkens the backdrop color to reflect the source color. Painting with white
     produces no change."
    """
    HARD_LIGHT = Name("HardLight")
    """
    "Multiplies or screens the colors, depending on the source color value. The effect
    is similar to shining a harsh spotlight on the backdrop."
    """
    SOFT_LIGHT = Name("SoftLight")
    """
    "Darkens or lightens the colors, depending on the source color value. The effect is
    similar to shining a diffused spotlight on the backdrop."
    """
    DIFFERENCE = Name("Difference")
    '''"Subtracts the darker of the two constituent colors from the lighter color."'''
    EXCLUSION = Name("Exclusion")
    """
    "Produces an effect similar to that of the Difference mode but lower in contrast.
    Painting with white inverts the backdrop color; painting with black produces no
    change."
    """
    HUE = Name("Hue")
    """
    "Creates a color with the hue of the source color and the saturation and luminosity
    of the backdrop color."
    """
    SATURATION = Name("Saturation")
    """
    "Creates a color with the saturation of the source color and the hue and luminosity
    of the backdrop color. Painting with this mode in an area of the backdrop that is
    a pure gray (no saturation) produces no change."
    """
    COLOR = Name("Color")
    """
    "Creates a color with the hue and saturation of the source color and the luminosity
    of the backdrop color. This preserves the gray levels of the backdrop and is
    useful for coloring monochrome images or tinting color images."
    """
    LUMINOSITY = Name("Luminosity")
    """
    "Creates a color with the luminosity of the source color and the hue and saturation
    of the backdrop color. This produces an inverse effect to that of the Color mode."
    """


class CompositingOperation(CoerciveEnum):
    "An enumeration of Porter-Duff compositing operations."

    CLEAR = Name("Clear")
    """ Draw nothing """

    SOURCE = Name("Source")
    """ Draw the source only """

    DESTINATION = Name("Destination")
    """ Draw the destination only """

    SOURCE_OVER = Name("SourceOver")
    """The source is drawn over the destination (backdrop)."""

    DESTINATION_OVER = Name("DestinationOver")
    """The destination (backdrop) is drawn over the source."""

    SOURCE_IN = Name("SourceIn")
    """Only the part of the source that overlaps with the destination is drawn. The rest is discarded."""

    DESTINATION_IN = Name("DestinationIn")
    """Only the part of the destination that overlaps with the source is drawn. The rest is discarded."""

    SOURCE_OUT = Name("SourceOut")
    """Only the part of the source that does not overlap the destination is drawn."""

    DESTINATION_OUT = Name("DestinationOut")
    """Only the part of the destination that does not overlap the source is drawn."""

    SOURCE_ATOP = Name("SourceAtop")
    """The part of the source that overlaps the destination is drawn over the destination. The rest of the source is discarded."""

    DESTINATION_ATOP = Name("DestinationAtop")
    """The part of the destination that overlaps the source is drawn over the source. The rest of the destination is discarded."""

    XOR = Name("XOR")
    """Only the parts of the source and destination that do not overlap are drawn."""


class AnnotationFlag(CoerciveIntEnum):
    INVISIBLE = 1
    """
    If set, do not display the annotation if it does not belong to one of the
    standard annotation types and no annotation handler is available.
    """
    HIDDEN = 2
    "If set, do not display or print the annotation or allow it to interact with the user"
    PRINT = 4
    "If set, print the annotation when the page is printed."
    NO_ZOOM = 8
    "If set, do not scale the annotation’s appearance to match the magnification of the page."
    NO_ROTATE = 16
    "If set, do not rotate the annotation’s appearance to match the rotation of the page."
    NO_VIEW = 32
    "If set, do not display the annotation on the screen or allow it to interact with the user"
    READ_ONLY = 64
    """
    If set, do not allow the annotation to interact with the user.
    The annotation may be displayed or printed but should not respond to mouse clicks.
    """
    LOCKED = 128
    """
    If set, do not allow the annotation to be deleted or its properties
    (including position and size) to be modified by the user.
    """
    TOGGLE_NO_VIEW = 256
    "If set, invert the interpretation of the NoView flag for certain events."
    LOCKED_CONTENTS = 512
    "If set, do not allow the contents of the annotation to be modified by the user."


class AnnotationName(CoerciveEnum):
    "The name of an icon that shall be used in displaying the annotation"

    NOTE = Name("Note")
    COMMENT = Name("Comment")
    HELP = Name("Help")
    PARAGRAPH = Name("Paragraph")
    NEW_PARAGRAPH = Name("NewParagraph")
    INSERT = Name("Insert")


class FileAttachmentAnnotationName(CoerciveEnum):
    "The name of an icon that shall be used in displaying the annotation"

    PUSH_PIN = Name("PushPin")
    GRAPH_PUSH_PIN = Name("GraphPushPin")
    PAPERCLIP_TAG = Name("PaperclipTag")


class IntersectionRule(CoerciveEnum):
    """
    An enumeration representing the two possible PDF intersection rules.

    The intersection rule is used by the renderer to determine which points are
    considered to be inside the path and which points are outside the path. This
    primarily affects fill rendering and clipping paths.
    """

    NONZERO = "nonzero"
    """
    "The nonzero winding number rule determines whether a given point is inside a path
    by conceptually drawing a ray from that point to infinity in any direction and
    then examining the places where a segment of the path crosses the ray. Starting
    with a count of 0, the rule adds 1 each time a path segment crosses the ray from
    left to right and subtracts 1 each time a segment crosses from right to left.
    After counting all the crossings, if the result is 0, the point is outside the
    path; otherwise, it is inside."
    """
    EVENODD = "evenodd"
    """
    "An alternative to the nonzero winding number rule is the even-odd rule. This rule
    determines whether a point is inside a path by drawing a ray from that point in
    any direction and simply counting the number of path segments that cross the ray,
    regardless of direction. If this number is odd, the point is inside; if even, the
    point is outside. This yields the same results as the nonzero winding number rule
    for paths with simple shapes, but produces different results for more complex
    shapes."
    """


class PathPaintRule(CoerciveEnum):
    """
    An enumeration of the PDF drawing directives that determine how the renderer should
    paint a given path.
    """

    # the auto-close paint rules are omitted here because it's easier to just emit
    # close operators when appropriate, programmatically
    STROKE = "S"
    '''"Stroke the path."'''

    FILL_NONZERO = "f"
    """
    "Fill the path, using the nonzero winding number rule to determine the region to
    fill. Any subpaths that are open are implicitly closed before being filled."
    """

    FILL_EVENODD = "f*"
    """
    "Fill the path, using the even-odd rule to determine the region to fill. Any
    subpaths that are open are implicitly closed before being filled."
    """

    STROKE_FILL_NONZERO = "B"
    """
    "Fill and then stroke the path, using the nonzero winding number rule to determine
    the region to fill. This operator produces the same result as constructing two
    identical path objects, painting the first with `FILL_NONZERO` and the second with
    `STROKE`."
    """

    STROKE_FILL_EVENODD = "B*"
    """
    "Fill and then stroke the path, using the even-odd rule to determine the region to
    fill. This operator produces the same result as `STROKE_FILL_NONZERO`, except that
    the path is filled as if with `FILL_EVENODD` instead of `FILL_NONZERO`."
    """

    DONT_PAINT = "n"
    """
    "End the path object without filling or stroking it. This operator is a
    path-painting no-op, used primarily for the side effect of changing the current
    clipping path."
    """

    AUTO = "auto"
    """
    Automatically determine which `PathPaintRule` should be used.

    PaintedPath will select one of the above `PathPaintRule`s based on the resolved
    set/inherited values of its style property.
    """


class ClippingPathIntersectionRule(CoerciveEnum):
    "An enumeration of the PDF drawing directives that define a path as a clipping path."

    NONZERO = "W"
    """
    "The nonzero winding number rule determines whether a given point is inside a path
    by conceptually drawing a ray from that point to infinity in any direction and
    then examining the places where a segment of the path crosses the ray. Starting
    with a count of 0, the rule adds 1 each time a path segment crosses the ray from
    left to right and subtracts 1 each time a segment crosses from right to left.
    After counting all the crossings, if the result is 0, the point is outside the
    path; otherwise, it is inside."
    """
    EVENODD = "W*"
    """
    "An alternative to the nonzero winding number rule is the even-odd rule. This rule
    determines whether a point is inside a path by drawing a ray from that point in
    any direction and simply counting the number of path segments that cross the ray,
    regardless of direction. If this number is odd, the point is inside; if even, the
    point is outside. This yields the same results as the nonzero winding number rule
    for paths with simple shapes, but produces different results for more complex
    shapes."""


class StrokeCapStyle(CoerciveIntEnum):
    """
    An enumeration of values defining how the end of a stroke should be rendered.

    This affects the ends of the segments of dashed strokes, as well.
    """

    BUTT = 0
    """
    "The stroke is squared off at the endpoint of the path. There is no projection
    beyond the end of the path."
    """
    ROUND = 1
    """
    "A semicircular arc with a diameter equal to the line width is drawn around the
    endpoint and filled in."
    """
    SQUARE = 2
    """
    "The stroke continues beyond the endpoint of the path for a distance equal to half
    the line width and is squared off."
    """


class StrokeJoinStyle(CoerciveIntEnum):
    """
    An enumeration of values defining how the corner joining two path components should
    be rendered.
    """

    MITER = 0
    """
    "The outer edges of the strokes for the two segments are extended until they meet at
    an angle, as in a picture frame. If the segments meet at too sharp an angle
    (as defined by the miter limit parameter), a bevel join is used instead."
    """
    ROUND = 1
    """
    "An arc of a circle with a diameter equal to the line width is drawn around the
    point where the two segments meet, connecting the outer edges of the strokes for
    the two segments. This pieslice-shaped figure is filled in, pro- ducing a rounded
    corner."
    """
    BEVEL = 2
    """
    "The two segments are finished with butt caps and the resulting notch beyond the
    ends of the segments is filled with a triangle."
    """


class PDFStyleKeys(Enum):
    "An enumeration of the graphics state parameter dictionary keys."

    FILL_ALPHA = Name("ca")
    BLEND_MODE = Name("BM")  # shared between stroke and fill
    STROKE_ALPHA = Name("CA")
    STROKE_ADJUSTMENT = Name("SA")
    STROKE_WIDTH = Name("LW")
    STROKE_CAP_STYLE = Name("LC")
    STROKE_JOIN_STYLE = Name("LJ")
    STROKE_MITER_LIMIT = Name("ML")
    STROKE_DASH_PATTERN = Name("D")  # array of array, number, e.g. [[1 1] 0]
    SOFT_MASK = Name("SMask")


class Corner(CoerciveEnum):
    TOP_RIGHT = "TOP_RIGHT"
    TOP_LEFT = "TOP_LEFT"
    BOTTOM_RIGHT = "BOTTOM_RIGHT"
    BOTTOM_LEFT = "BOTTOM_LEFT"


class FontDescriptorFlags(Flag):
    """An enumeration of the flags for the unsigned 32-bit integer entry in the font descriptor specifying various
    characteristics of the font. Bit positions are numbered from 1 (low-order) to 32 (high-order).
    """

    FIXED_PITCH = 0x0000001
    """
    "All glyphs have the same width (as opposed to proportional or
    variable-pitch fonts, which have different widths."
    """

    SYMBOLIC = 0x0000004
    """
    "Font contains glyphs outside the Adobe standard Latin character set.
    This flag and the Nonsymbolic flag shall not both be set or both be clear."
    """

    ITALIC = 0x0000040
    """
    "Glyphs have dominant vertical strokes that are slanted."
    """

    FORCE_BOLD = 0x0040000
    """
    "The flag shall determine whether bold glyphs shall be painted with extra pixels even at very
    small text sizes by a conforming reader. If set, features of bold glyphs may be thickened at
    small text sizes."
    """


class AccessPermission(IntFlag):
    "Permission flags will translate as an integer on the encryption dictionary"

    PRINT_LOW_RES = 0b000000000100
    "Print the document"

    MODIFY = 0b000000001000
    "Modify the contents of the document"

    COPY = 0b000000010000
    "Copy or extract text and graphics from the document"

    ANNOTATION = 0b000000100000
    "Add or modify text annotations"

    FILL_FORMS = 0b000100000000
    "Fill in existing interactive form fields"

    COPY_FOR_ACCESSIBILITY = 0b001000000000
    "Extract text and graphics in support of accessibility to users with disabilities"

    ASSEMBLE = 0b010000000000
    "Insert, rotate or delete pages and create bookmarks or thumbnail images"

    PRINT_HIGH_RES = 0b100000000000
    "Print document at the highest resolution"

    @classmethod
    def all(cls) -> int:
        "All flags enabled"
        result = 0
        for permission in list(AccessPermission):
            access_permission = permission
            result = result | access_permission.value
        return result

    @classmethod
    def none(cls) -> int:
        "All flags disabled"
        return 0


class EncryptionMethod(Enum):
    "Algorithm to be used to encrypt the document"

    NO_ENCRYPTION = 0
    RC4 = 1
    AES_128 = 2
    AES_256 = 3


class TextDirection(CoerciveEnum):
    "Text rendering direction for text shaping"

    LTR = intern("LTR")
    "left to right"

    RTL = intern("RTL")
    "right to left"

    TTB = intern("TTB")
    "top to bottom"

    BTT = intern("BTT")
    "bottom to top"


class OutputIntentSubType(CoerciveEnum):
    "Definition for Output Intent Subtypes"

    PDFX = Name("GTS_PDFX")
    "PDF/X-1a which is based upon CMYK processing"

    PDFA = Name("GTS_PDFA1")
    "PDF/A (ISO 19005) standard to produce RGB output"

    ISOPDF = Name("ISO_PDFE1")
    "ISO_PDFE1 PDF/E standards (ISO 24517, all parts)"


class PageLabelStyle(CoerciveEnum):
    "Style of the page label"

    NUMBER = intern("D")
    "decimal arabic numerals"

    UPPER_ROMAN = intern("R")
    "uppercase roman numerals"

    LOWER_ROMAN = intern("r")
    "lowercase roman numerals"

    UPPER_LETTER = intern("A")
    "uppercase letters A to Z, AA to ZZ, AAA to ZZZ and so on"

    LOWER_LETTER = intern("a")
    "uppercase letters a to z, aa to zz, aaa to zzz and so on"

    NONE = None
    "no label"


class Duplex(CoerciveEnum):
    "The paper handling option that shall be used when printing the file from the print dialog."

    SIMPLEX = Name("Simplex")
    "Print single-sided"

    DUPLEX_FLIP_SHORT_EDGE = Name("DuplexFlipShortEdge")
    "Duplex and flip on the short edge of the sheet"

    DUPLEX_FLIP_LONG_EDGE = Name("DuplexFlipLongEdge")
    "Duplex and flip on the long edge of the sheet"


class PageBoundaries(CoerciveEnum):
    ART_BOX = Name("ArtBox")
    BLEED_BOX = Name("BleedBox")
    CROP_BOX = Name("CropBox")
    MEDIA_BOX = Name("MediaBox")
    TRIM_BOX = Name("TrimBox")


class PageOrientation(CoerciveEnum):
    PORTRAIT = intern("P")
    LANDSCAPE = intern("L")

    @classmethod
    def coerce(  # pyright: ignore[reportIncompatibleMethodOverride]
        cls, value: Union["PageOrientation", str], case_sensitive: bool = False
    ) -> "PageOrientation":
        if isinstance(value, str):
            value = value.upper()
        return super(cls, cls).coerce(
            value, case_sensitive  # pyright: ignore[reportArgumentType]
        )


class PDFResourceType(Enum):
    EXT_G_STATE = intern("ExtGState")
    COLOR_SPACE = intern("ColorSpace")
    PATTERN = intern("Pattern")
    SHADING = intern("Shading")
    X_OBJECT = intern("XObject")
    FONT = intern("Font")
    PROC_SET = intern("ProcSet")
    PROPERTIES = intern("Properties")


class GradientUnits(CoerciveEnum):
    "Specifies the coordinate system for gradients."

    OBJECT_BOUNDING_BOX = "objectBoundingBox"
    " Coordinates are expressed as fractions of the painted object's bounding box (0..1 in each axis)."

    USER_SPACE_ON_USE = "userSpaceOnUse"
    " Coordinates are in the current page space."


class GradientSpreadMethod(CoerciveEnum):
    "Specifies how to fill the area outside the gradient's start and end points."

    PAD = "pad"
    " The color at the start or end of the gradient is extended to fill the area before or after the gradient."

    REFLECT = "reflect"
    " The gradient pattern is repeated in reverse order (mirrored) to fill the area before or after the gradient."

    REPEAT = "repeat"
    " The gradient pattern is repeated in the same order to fill the area before or after the gradient."


class DocumentCompliance(Enum):
    """
    Type of compliance enforcement that can be applied to a document.
    Limited to PDF/A at the moment, but extendable to other standards like:
        - PDF/E (Engineering PDFs)
        - PDF/UA (PDF Universal Accessibility)
        - PDF/X (Graphics Exchange PDFs)
    """

    PDFA_1B = ("PDFA", 1, "B")
    PDFA_2B = ("PDFA", 2, "B")
    PDFA_2U = ("PDFA", 2, "U")
    PDFA_3B = ("PDFA", 3, "B")
    PDFA_3U = ("PDFA", 3, "U")
    PDFA_4 = ("PDFA", 4, None)
    PDFA_4E = ("PDFA", 4, "E")
    PDFA_4F = ("PDFA", 4, "F")

    @property
    def profile(self) -> str:
        return str(self.value[0])

    @property
    def part(self) -> int:
        return int(self.value[1])

    @property
    def conformance(self) -> Optional[str]:
        return str(self.value[2]) if self.value[2] is not None else None

    @property
    def label(self) -> str:
        profile = "PDF/A" if self.profile == "PDFA" else self.profile
        return f"{profile}-{self.part}{self.conformance if self.conformance else ''}"

    def __str__(self) -> str:
        return (
            f"{self.profile}_{self.part}{self.conformance if self.conformance else ''}"
        )

    @classmethod
    def coerce(cls, value: Union["DocumentCompliance", str]) -> "DocumentCompliance":
        if isinstance(value, cls):
            return value
        if isinstance(value, str):
            key = value.upper()
            for m in cls:
                if m.name.upper() == key:  # PDFA_2U
                    return m
                if m.label.upper() == key:  # PDF/A-2U
                    return m
        raise ValueError(f"Cannot coerce {value!r} to {cls.__name__}")


class AssociatedFileRelationship(CoerciveEnum):
    """Represents the association between an embedded file and the content on the PDF"""

    SOURCE = intern("Source")
    "The file is the original source material of the content"

    DATA = intern("Data")
    """
    The file has the information used to produce the associated object.
    e.g.: the data used to produce a table or a graph
    """

    ALTERNATIVE = intern("Alternative")
    "The file has an alternative representation of the content"

    SUPPLEMENT = intern("Supplement")
    """
    The file has a supplemental representation of the original source
    or data that may be more easily consumable
    """

    ENCRYPTED_PAYLOAD = intern("EncryptedPayload")
    """
    The file is an encrypted payload document that should be displayed
    to the user if the PDF processor has the cryptographic filter
    needed to decrypt the document
    """

    FORM_DATA = intern("FormData")
    "The file has the data associated with the interactive form in this document"

    SCHEMA = intern("Schema")
    "The file is a schema definition for the associated object"

    UNSPECIFIED = intern("Unspecified")
    """
    Shall be used when the relationship is not known
    or cannot be described using one of the other values
    """



"""
Utilities to perform encryption following the PDF standards.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.

Usage documentation at: <https://py-pdf.github.io/fpdf2/Encryption.html>
"""

# pyright: reportPossiblyUnboundVariable=false

import hashlib
import logging
import math
import stringprep
import unicodedata
from binascii import hexlify
from codecs import BOM_UTF16_BE
from os import urandom
from typing import TYPE_CHECKING, Callable, Iterator, Optional, Sequence, Type, Union

from .enums import AccessPermission, EncryptionMethod
from .errors import FPDFException
from .syntax import (
    Name,
    PDFObject,
    PDFString,
    build_obj_dict,
    create_dictionary_string as pdf_dict,
)

# try to use cryptography for AES encryption
try:
    from cryptography.hazmat.primitives.ciphers import Cipher, modes
    from cryptography.hazmat.primitives.ciphers.algorithms import AES128, AES256
    from cryptography.hazmat.primitives.padding import PKCS7

    import_error = None
except ImportError as error:
    import_error = error

if TYPE_CHECKING:
    from .fpdf import FPDF

LOGGER = logging.getLogger(__name__)


class ARC4:
    """
    This is a simplified version of the ARC4 (alleged RC4) algorithm,
    created based on the following sources:
    * Wikipedia article on RC4
    * github.com/manojpandey/rc4 (MIT License)
    * http://people.csail.mit.edu/rivest/pubs/RS14.pdf

    Having this ARC4 implementation makes it possible to have basic
    encryption functions without additional dependencies
    """

    MOD = 256

    def KSA(self, key: bytes) -> list[int]:
        key_length = len(key)
        S = list(range(self.MOD))
        j = 0
        for i in range(self.MOD):
            j = (j + S[i] + key[i % key_length]) % self.MOD
            S[i], S[j] = S[j], S[i]
        return S

    def PRGA(self, S: list[int]) -> Iterator[int]:
        i = 0
        j = 0
        while True:
            i = (i + 1) % self.MOD
            j = (j + S[i]) % self.MOD
            S[i], S[j] = S[j], S[i]
            K = S[(S[i] + S[j]) % self.MOD]
            yield K

    def encrypt(self, key: bytes, text: bytes | bytearray) -> list[int]:
        keystream = self.PRGA(self.KSA(key))
        res: list[int] = []
        for c in text:
            res.append(c ^ next(keystream))
        return res


class CryptFilter:
    """Represents one crypt filter, listed under CF inside the encryption dictionary"""

    def __init__(self, mode: str, length: int) -> None:
        super().__init__()
        self.type = Name("CryptFilter")
        self.c_f_m = Name(mode)
        self.length = int(length / 8)

    def serialize(self) -> str:
        obj_dict = build_obj_dict({key: getattr(self, key) for key in dir(self)})
        return pdf_dict(obj_dict)


class EncryptionDictionary(PDFObject):
    """
    This class represents an encryption dictionary
    PDF 32000 reference - Table 20
    The PDF trailer must reference this object (/Encrypt)
    """

    def __init__(self, security_handler: "StandardSecurityHandler") -> None:
        super().__init__()
        self.filter = Name("Standard")
        self.length = security_handler.key_length
        self.r = security_handler.revision
        self.o = f"<{security_handler.o.upper()}>"
        self.u = f"<{security_handler.u.upper()}>"
        if security_handler.revision == 6:
            self.o_e = f"<{security_handler.oe.upper()}>"
            self.u_e = f"<{security_handler.ue.upper()}>"
            self.perms = f"<{security_handler.perms.upper()}>"
        self.v = security_handler.version
        self.p = int32(security_handler.access_permission)
        if not security_handler.encrypt_metadata:
            self.encrypt_metadata = "false"
        if security_handler.cf:
            self.c_f = pdf_dict({"/StdCF": security_handler.cf.serialize()})
        if security_handler.encryption_method == EncryptionMethod.NO_ENCRYPTION:
            self.stm_f = Name("Identity")  # crypt filter for streams
            self.str_f = Name("Identity")  # crypt filter for strings
        else:
            self.stm_f = Name("StdCF")  # crypt filter for streams
            self.str_f = Name("StdCF")  # crypt filter for strings


class StandardSecurityHandler:
    """
    This class is referenced in the main PDF class and is used to handle all encryption functions
        * Calculate password and hashes
        * Provide encrypt method to be called by stream and strings
        * Set the access permissions on the document
    """

    DEFAULT_PADDING = (
        b"(\xbfN^Nu\x8aAd\x00NV\xff\xfa\x01\x08..\x00\xb6\xd0h>\x80/\x0c\xa9\xfedSiz"
    )

    def __init__(
        self,
        fpdf: "FPDF",
        owner_password: str,
        user_password: Optional[str] = None,
        permission: int | AccessPermission = AccessPermission.all(),
        encryption_method: EncryptionMethod = EncryptionMethod.RC4,
        encrypt_metadata: bool = False,
    ):
        self.fpdf = fpdf
        self.access_permission = 0b11111111111111111111000011000000 | permission
        self.owner_password = owner_password
        self.user_password = user_password if user_password else ""
        self.encryption_method = encryption_method
        self.cf = None
        self.key_length = 128

        if import_error and self.encryption_method in (
            EncryptionMethod.AES_128,
            EncryptionMethod.AES_256,
        ):
            raise EnvironmentError(
                "cryptography module not available"
                " - Try: 'pip install cryptography' or use RC4 encryption method"
                f" - Import error was: {import_error}"
            )
        if self.encryption_method == EncryptionMethod.AES_128:
            self.version = 4
            self.revision = 4
            fpdf._set_min_pdf_version("1.6")  # pyright: ignore[reportPrivateUsage]
            self.cf = CryptFilter(mode="AESV2", length=self.key_length)
        elif self.encryption_method == EncryptionMethod.AES_256:
            self.version = 5
            self.revision = 6
            fpdf._set_min_pdf_version("2.0")  # pyright: ignore[reportPrivateUsage]
            self.key_length = 256
            self.cf = CryptFilter(mode="AESV3", length=self.key_length)
        elif self.encryption_method == EncryptionMethod.NO_ENCRYPTION:
            self.version = 4
            self.revision = 4
            fpdf._set_min_pdf_version("1.6")  # pyright: ignore[reportPrivateUsage]
            self.cf = CryptFilter(mode="V2", length=self.key_length)
        else:
            self.version = 2
            self.revision = 3
            fpdf._set_min_pdf_version("1.5")  # pyright: ignore[reportPrivateUsage]
            # not including crypt filter because it's only required on V=4
            # if needed, it would be CryptFilter(mode=V2)

        self.encrypt_metadata = encrypt_metadata

    def generate_passwords(self, file_id: str) -> None:
        """File_id is the first hash of the PDF file id"""
        self.file_id = file_id
        self.info_id = file_id[1:33]
        if self.revision == 6:
            self.k = self.get_random_bytes(32)
            self.generate_user_password_rev6()
            self.generate_owner_password_rev6()
            self.generate_perms_rev6()
        else:
            self.o = self.generate_owner_password()
            self.k = self.generate_encryption_key()
            self.u = self.generate_user_password()

    def get_encryption_obj(self) -> EncryptionDictionary:
        """Return an encryption dictionary"""
        return EncryptionDictionary(self)

    def encrypt(
        self, text: Union[str, bytearray, bytes], obj_id: int
    ) -> Union[str, bytes]:
        """Method invoked by PDFObject and PDFContentStream to encrypt strings and streams"""
        LOGGER.debug("Encrypting %s", text)
        return (
            self.encrypt_stream(text, obj_id)
            if isinstance(text, (bytes, bytearray))
            else self.encrypt_string(text, obj_id)
        )

    def encrypt_string(self, string: str, obj_id: int) -> str:
        if self.encryption_method == EncryptionMethod.NO_ENCRYPTION:
            return PDFString(string, encrypt=False).serialize()
        LOGGER.debug("Encrypting string: %s", string)
        try:
            string.encode("latin-1")
            return f"<{bytes(self.encrypt_bytes(string.encode('latin-1'), obj_id)).hex().upper()}>"
        except UnicodeEncodeError:
            return f'<{hexlify(bytearray(self.encrypt_bytes(BOM_UTF16_BE + string.encode("utf-16-be"), obj_id))).decode("latin-1")}>'

    def encrypt_stream(self, stream: bytes | bytearray, obj_id: int) -> bytes:
        if self.encryption_method == EncryptionMethod.NO_ENCRYPTION:
            return bytes(stream)
        return bytes(self.encrypt_bytes(stream, obj_id))

    def is_aes_algorithm(self) -> bool:
        return self.encryption_method in (
            EncryptionMethod.AES_128,
            EncryptionMethod.AES_256,
        )

    def encrypt_bytes(self, data: bytes | bytearray, obj_id: int) -> Sequence[int]:
        """
        PDF32000 reference - Algorithm 1: Encryption of data using the RC4 or AES algorithms
        Append object ID and generation ID to the key and encrypt the data
        Generation ID is fixed as 0. Will need to revisit if the application start changing generation ID
        """
        h = hashlib.new("md5", usedforsecurity=False)
        h.update(self.k)
        h.update(
            (obj_id & 0xFFFFFF).to_bytes(3, byteorder="little", signed=False)
        )  # object id
        h.update(
            (0 & 0xFFFF).to_bytes(2, byteorder="little", signed=False)
        )  # generation id
        if self.is_aes_algorithm():
            h.update(bytes([0x73, 0x41, 0x6C, 0x54]))  # add salt (sAlT) for AES
        key = h.digest()

        if self.is_aes_algorithm():
            return self.encrypt_AES_cryptography(key, data)
        return ARC4().encrypt(key, data)

    def encrypt_AES_cryptography(
        self, key: bytes, data: bytes | bytearray
    ) -> bytearray:
        """Encrypts an array of bytes using AES algorithms (AES 128 or AES 256)"""
        iv = bytearray(self.get_random_bytes(16))
        padder = PKCS7(128).padder()
        padded_data = padder.update(data)
        padded_data += padder.finalize()
        cipher = (
            Cipher(AES128(key), modes.CBC(iv))
            if self.encryption_method == EncryptionMethod.AES_128
            else Cipher(AES256(self.k), modes.CBC(iv))
        )
        encryptor = cipher.encryptor()
        data = encryptor.update(padded_data) + encryptor.finalize()
        iv.extend(data)
        return iv

    @classmethod
    def get_random_bytes(cls: Type["StandardSecurityHandler"], size: int) -> bytes:
        """
        https://docs.python.org/3/library/os.html#os.urandom
        os.urandom will use OS-specific sources to generate random bytes
        suitable for cryptographic use
        """
        return urandom(size)

    @classmethod
    def prepare_string(cls: Type["StandardSecurityHandler"], string: str) -> bytes:
        """
        PDF2.0 - ISO 32000-2:2020
        All passwords for revision 6 shall be based on Unicode. Preprocessing of a user-provided password
        consists first of normalizing its representation by applying the "SASLPrep" profile (Internet RFC 4013)
        of the "stringprep" algorithm (Internet RFC 3454) to the supplied password using the Normalize and BiDi
        options. Next, the password string shall be converted to UTF-8 encoding, and then truncated to the
        first 127 bytes if the string is longer than 127 bytes

        Python offers a stringprep module with the tables mapped in methods
        """

        # Mapping
        def char_map(char: str) -> str:
            if not char:
                return ""
            # Commonly mapped to nothing
            if stringprep.in_table_b1(char):
                return ""
            # Map non-ascii space characters to space
            if stringprep.in_table_c12(char):
                return "\u0020"
            return char

        if len(string) < 1:
            return bytes()

        prepared_string = "".join(char_map(c) for c in string)

        # Normalization - applies Unicode normalization form KC
        prepared_string = unicodedata.ucd_3_2_0.normalize("NFKC", prepared_string)

        # Prohibited output - RCF4013 2.3
        def is_prohibited(char: str) -> bool:
            return (
                stringprep.in_table_c12(char)  # Non-ASCII space characters
                or stringprep.in_table_c21_c22(char)  # Control characters
                or stringprep.in_table_c3(char)  # Private use
                or stringprep.in_table_c4(char)  # Non-character code points
                or stringprep.in_table_c5(char)  # Surrogate codes
                or stringprep.in_table_c6(char)  # Inappropriate for plain text
                or stringprep.in_table_c7(
                    char
                )  # Inappropriate for canonical representation
                or stringprep.in_table_c8(
                    char
                )  # Change display properties or are deprecated
                or stringprep.in_table_c9(char)  # Tagging characters
            )

        for char in prepared_string:
            if is_prohibited(char):
                raise FPDFException(
                    f"The password {string} contains prohibited characters"
                )

        # Bidirectional characters
        def has_character(string: str, fun: Callable[[str], bool]) -> bool:
            return any(fun(char) for char in string)

        if has_character(prepared_string, stringprep.in_table_d1):
            # If a string contains any RandALCat character, the string MUST NOT contain any LCat character.
            if has_character(prepared_string, stringprep.in_table_d2):
                raise FPDFException(
                    f"The password {string} contains invalid bidirectional characters."
                )
            # If a string contains any RandALCat character, a RandALCat character MUST be the first character
            # of the string, and a RandALCat character MUST be the last character of the string.
            if not (
                stringprep.in_table_d1(prepared_string[0])
                and stringprep.in_table_d1(prepared_string[-1])
            ):
                raise FPDFException(
                    f"The password {string} contains invalid bidirectional characters."
                )

        if len(prepared_string) > 127:
            prepared_string = prepared_string[:127]

        return prepared_string.encode("UTF-8")

    def padded_password(self, password: str) -> bytearray:
        """
        PDF32000 reference - Algorithm 2: Computing an encryption key
        Step (a) - Add the default padding at the end of provided password to make it 32 bit long
        """
        if len(password) > 32:
            password = password[:32]
        p = bytearray(password.encode("latin1"))
        p.extend(self.DEFAULT_PADDING[: (32 - len(p))])
        return p

    def generate_owner_password(self) -> str:
        """
        PDF32000 reference - Algorithm 3: Computing the encryption dictionary's O (owner password) value
        The security handler is only using revision 3 or 4, so the legacy r2 version is not implemented here
        """
        m = bytes(self.padded_password(self.owner_password))
        for _ in range(51):
            m = md5(m)
        rc4key = m[: (math.ceil(self.key_length / 8))]
        result = self.padded_password(self.user_password)
        for i in range(20):
            new_key: list[int] = []
            for k in rc4key:
                new_key.append(k ^ i)
            result = bytearray(ARC4().encrypt(bytes(new_key), result))
        return bytes(result).hex()

    def generate_user_password(self) -> str:
        """
        PDF32000 reference - Algorithm 5: Computing the encryption dictionary's U (user password) value
        The security handler is only using revision 3 or 4, so the legacy r2 version is not implemented here
        """
        m = hashlib.new("md5", usedforsecurity=False)
        m.update(bytearray(self.DEFAULT_PADDING))
        m.update(bytes.fromhex(self.info_id))
        result = bytearray(m.digest())
        key = self.k
        for i in range(20):
            new_key: list[int] = []
            for k in key:
                new_key.append(k ^ i)
            result = bytearray(ARC4().encrypt(bytes(new_key), result))
        result.extend(
            (result[x] ^ self.DEFAULT_PADDING[x]) for x in range(16)
        )  # add 16 bytes of random padding
        return bytes(result).hex()

    @classmethod
    def compute_hash(
        cls: Type["StandardSecurityHandler"],
        input_password: bytes | bytearray,
        salt: bytes,
        user_key: bytes | bytearray = bytearray(),
    ) -> bytes:
        """
        Algorithm 2B - section 7.6.4.3.4 of the ISO 32000-2:2020
        Applied on Security handlers revision 6
        """
        k = hashlib.sha256(input_password + salt + user_key).digest()
        round_number = 0
        while True:
            round_number += 1
            k1 = input_password + k + user_key
            # Step (a + b)
            cipher = Cipher(AES128(k[:16]), modes.CBC(k[16:32]))
            encryptor = cipher.encryptor()
            e = encryptor.update(k1 * 64) + encryptor.finalize()
            # Step (c)
            # remainder = int.from_bytes(e[:16], byteorder="big") % 3
            remainder = sum(e[:16]) % 3
            # Step (d)
            if remainder == 0:
                k = hashlib.sha256(e).digest()
            elif remainder == 1:
                k = hashlib.sha384(e).digest()
            else:
                k = hashlib.sha512(e).digest()
            # Step (e)
            if round_number >= 64 and e[-1] <= round_number - 32:
                break

        return k[:32]

    def generate_user_password_rev6(self) -> None:
        """
        Generating the U (user password) and UE (user encryption)
        for security handlers of revision 6
        Algorithm 8 - Section 7.6.4.4.7 of the ISO 32000-2:2020
        """
        user_password: bytes | bytearray = self.prepare_string(self.user_password)
        if not user_password:
            user_password = bytearray()
        user_validation_salt = self.get_random_bytes(8)
        user_key_salt = self.get_random_bytes(8)
        u = (
            self.compute_hash(input_password=user_password, salt=user_validation_salt)
            + user_validation_salt
            + user_key_salt
        )
        self.u = u.hex()

        key = self.compute_hash(input_password=user_password, salt=user_key_salt)
        cipher = Cipher(AES256(key), modes.CBC(b"\x00" * 16))
        encryptor = cipher.encryptor()
        ue = encryptor.update(self.k) + encryptor.finalize()
        self.ue = ue.hex()

    def generate_owner_password_rev6(self) -> None:
        """
        Generating the O (owner password) and OE (owner encryption)
        for security handlers of revision 6
        Algorithm 9 - Section 7.6.4.4.8 of the ISO 32000-2:2020
        """
        owner_password = self.prepare_string(self.owner_password)
        if not owner_password:
            raise FPDFException(f"Invalid owner password {self.owner_password}")
        owner_validation_salt = self.get_random_bytes(8)
        owner_key_salt = self.get_random_bytes(8)
        o = (
            self.compute_hash(
                input_password=owner_password,
                salt=owner_validation_salt,
                user_key=bytes.fromhex(self.u),
            )
            + owner_validation_salt
            + owner_key_salt
        )
        self.o = o.hex()

        key = self.compute_hash(
            input_password=owner_password,
            salt=owner_key_salt,
            user_key=bytes.fromhex(self.u),
        )

        cipher = Cipher(AES256(key), modes.CBC(b"\x00" * 16))
        encryptor = cipher.encryptor()
        oe = encryptor.update(self.k) + encryptor.finalize()
        self.oe = oe.hex()

    def generate_perms_rev6(self) -> None:
        """
        7.6.4.4.9 Algorithm 10: Computing the encryption dictionary’s Perms (permissions) value
        (Security handlers of revision 6) of the ISO 32000-2:2020
        """
        perms64b = 0xFFFFFFFF00000000 | self.access_permission
        encrypt_metadata = b"T" if self.encrypt_metadata else b"F"
        perms_input = (
            perms64b.to_bytes(8, byteorder="little", signed=False)
            + encrypt_metadata
            + b"adb"
            + self.get_random_bytes(4)
        )
        # nosemgrep: python.cryptography.security.insecure-cipher-mode-ecb.insecure-cipher-mode-ecb
        cipher = Cipher(AES256(self.k), modes.ECB())
        encryptor = cipher.encryptor()
        perms = encryptor.update(perms_input) + encryptor.finalize()
        self.perms = perms.hex()

    def generate_encryption_key(self) -> bytes:
        """
        PDF32000 reference
        Algorithm 2: Computing an encryption key
        """
        m = hashlib.new("md5", usedforsecurity=False)
        m.update(self.padded_password(self.user_password))
        m.update(bytes.fromhex(self.o))
        m.update(
            (self.access_permission & 0xFFFFFFFF).to_bytes(
                4, byteorder="little", signed=False
            )
        )
        m.update(bytes.fromhex(self.info_id))
        if self.encrypt_metadata is False and self.version == 4:
            m.update(bytes([0xFF, 0xFF, 0xFF, 0xFF]))
        result = m.digest()[: (math.ceil(self.key_length / 8))]
        for _ in range(50):
            result = md5(result)[: (math.ceil(self.key_length / 8))]
        return result


def md5(data: Union[bytes, bytearray]) -> bytes:
    h = hashlib.new("md5", usedforsecurity=False)
    h.update(data)
    return h.digest()


def int32(n: int) -> int:
    """convert long to signed 32 bit integer"""
    n = n & 0xFFFFFFFF
    return (n ^ 0x80000000) - 0x80000000



#!/usr/bin/env python3

# LICENSE: CC-0
# INSTALL: pip install agithub jinja2
# USAGE:
#   export GITHUB_TOKEN=...
#   ./build_contributors_html_page.py $org/$repo

# API DOC: https://developer.github.com/v3/issues/

import argparse
from http.client import HTTPConnection, HTTPException
import json
import logging
import os
import sys

from agithub.GitHub import GitHub
from agithub.base import IncompleteRequest
from jinja2 import Environment, FileSystemLoader

THIS_SCRIPT_PARENT_DIR = os.path.dirname(os.path.realpath(__file__))

EMOJI_KEY = {
    "a11y": "️♿️",
    "audio": "🔊",
    "blog": "📝",
    "bug": "🐛",
    "business": "💼",
    "code": "💻",
    "content": "🖋",
    "data": "🔣",
    "design": "🎨",
    "doc": "📖",
    "eventOrganizing": "📋",
    "example": "💡",
    "financial": "💵",
    "fundingFinding": "🔍",
    "ideas": "🤔",
    "infra": "🚇",
    "maintenance": "🚧",
    "mentoring": "🧑‍🏫",
    "platform": "📦",
    "plugin": "🔌",
    "projectManagement": "📆",
    "question": "💬",
    "research": "🔬",
    "review": "👀",
    "security": "🛡️",
    "talk": "📢",
    "test": "⚠️",
    "tool": "🔧",
    "translation": "🌍",
    "tutorial": "✅",
}


def main():
    if "GITHUB_TOKEN" not in os.environ:
        raise RuntimeError("Environment variable GITHUB_TOKEN must be defined")
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s - %(filename)s:%(lineno)d [%(levelname)s] %(message)s",
    )
    logging.getLogger("agithub.GitHub").setLevel(logging.DEBUG)
    args = parse_args()
    if args.debug:
        HTTPConnection.debuglevel = 2
    print("Reading .all-contributorsrc...")
    with open(
        f"{THIS_SCRIPT_PARENT_DIR}/../.all-contributorsrc", encoding="utf8"
    ) as json_file:
        allcontributors = json.load(json_file)["contributors"]
        contributors_per_login = {user["login"]: user for user in allcontributors}
    ag = GitHubAPIWrapper(token=os.environ["GITHUB_TOKEN"])
    org, repo = args.org_repo.split("/")
    print("Fetching all GitHub issues...")
    issues = ag.repos[org][repo].issues.get(state="all")
    print("Now retrieving each user location...")
    user_locations = {}
    for issue in issues:
        user_login = issue["user"]["login"]
        if user_login not in user_locations:
            contributions = []
            if user_login in contributors_per_login:
                contributions = contributors_per_login[user_login]["contributions"]
            user_locations[user_login] = {
                "contributions": contributions,
                "location": ag.users[user_login].get()["location"],
                "issues": 0,
                "pulls": 0,
            }
        if "pull_request" in issue:
            user_locations[user_login]["pulls"] += 1
        else:
            user_locations[user_login]["issues"] += 1
    print("Now adding remaining contributors from .all-contributorsrc...")
    for user in allcontributors:
        user_login = user["login"]
        if user_login not in user_locations:
            try:
                gh_user = ag.users[user_login].get()
                user_locations[user_login] = {
                    "contributions": user["contributions"],
                    "location": gh_user["location"],
                    "issues": 0,
                    "pulls": 0,
                }
            except HTTPException as error:
                print(error)
    print("Now generating the HTML page...")
    env = Environment(loader=FileSystemLoader(THIS_SCRIPT_PARENT_DIR), autoescape=True)
    template = env.get_template("contributors.html.jinja2")
    out_filepath = os.path.join(THIS_SCRIPT_PARENT_DIR, "contributors.html")
    with open(out_filepath, "w", encoding="utf-8") as out_file:
        out_file.write(
            template.render(
                org_repo=args.org_repo,
                user_locations=user_locations,
                EMOJI_KEY=EMOJI_KEY,
            )
        )


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False
    )
    parser.add_argument("org_repo")
    parser.add_argument("--debug", action="store_true")
    return parser.parse_args()


# Below are some utility methods to enhance the agithub library:


class GitHubAPIWrapper(GitHub):
    def __init__(self, *args, ignore_403s=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.ignore_403s = ignore_403s

    def __getattr__(self, key):
        return IncompleteRequestWrapper(self.client, self.ignore_403s).__getattr__(key)

    __getitem__ = __getattr__


class IncompleteRequestWrapper(IncompleteRequest):
    def __init__(self, client, ignore_403s):
        super().__init__(client)
        self.ignore_403s = ignore_403s

    def __getattr__(self, key):
        result = super().__getattr__(key)
        if key in self.client.http_methods:
            return HTTPRequester(result, self.ignore_403s)
        return result

    __getitem__ = __getattr__


class HTTPRequester:
    """
    Callable, providing:
    - auto pages fetching when result count is > 100
    - raise exceptions on HTTP errors
    """

    MAX_RESULTS_COUNT = 30

    def __init__(self, http_method_executer, ignore_403s):
        self.http_method_executer = http_method_executer
        self.ignore_403s = ignore_403s

    def __call__(self, *args, **kwargs):
        all_results, page = [], 1
        while len(all_results) % self.MAX_RESULTS_COUNT == 0:
            result = self._fetch(*args, page=page, **kwargs)
            if not isinstance(result, list):
                return result
            if len(result) == 0:
                return all_results
            all_results.extend(result)
            page += 1
        return all_results

    def _fetch(self, *args, **kwargs):
        http_code, response = self.http_method_executer(*args, **kwargs)
        if http_code == 403 and self.ignore_403s:
            print(f"HTTP {http_code}: {response['message']}", file=sys.stderr)
            return []
        if http_code != 200:
            url = self.http_method_executer.keywords["url"]
            raise HTTPException(f"[{url}] HTTP code: {http_code}: {response}")
        return response


if __name__ == "__main__":
    main()



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal
from fpdf.drawing import DeviceGray

HERE = Path(__file__).resolve().parent


def test_add_page_format(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica")
    for i in range(9):
        pdf.add_page(format=(210 * (1 - i / 10), 297 * (1 - i / 10)))
        pdf.cell(w=10, h=10, text=str(i))
    pdf.add_page(same=True)
    pdf.cell(w=10, h=10, text="9")
    assert_pdf_equal(pdf, HERE / "add_page_format.pdf", tmp_path)


def test_add_page_duration(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_font("Helvetica", size=120)
    pdf.add_page(duration=3)
    pdf.cell(text="Page 1")
    pdf.page_duration = 0.5
    pdf.add_page()
    pdf.cell(text="Page 2")
    pdf.add_page()
    pdf.cell(text="Page 3")
    assert_pdf_equal(pdf, HERE / "add_page_duration.pdf", tmp_path)


def test_break_or_add_page(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_auto_page_break(auto=True, margin=0)
    pdf.add_page()
    pdf.set_font("Helvetica", size=16)
    for i in range(1, 51):
        pdf.set_x(10)
        pdf.multi_cell(50, 10, f"Text {i} - Page {pdf.page}", 1, "C")
    pdf.page = 1
    pdf.set_xy(100, 10)
    for i in range(51, 101):
        pdf.set_x(100)
        pdf.multi_cell(50, 10, f"Text {i} - Page {pdf.page}", 1, "C")
    assert_pdf_equal(pdf, HERE / "break_or_add_page.pdf", tmp_path)


def test_break_or_add_page_with_different_draw_and_fill_color(tmp_path):
    class CustomHeader(fpdf.FPDF):
        def header(self):
            self.line_width = 0
            self.draw_color = DeviceGray(0.2)
            self.fill_color = DeviceGray(0.2)

    pdf = CustomHeader()
    pdf.set_auto_page_break(auto=True, margin=0)
    pdf.draw_color = DeviceGray(0.5)
    pdf.set_stretching(101)
    pdf.add_page()
    pdf.set_font("Helvetica", size=16)
    for i in range(1, 51):
        pdf.set_x(10)
        pdf.multi_cell(50, 10, f"Text {i} - Page {pdf.page}", 1, "C")
    pdf.page = 1
    pdf.set_xy(100, 10)
    for i in range(51, 101):
        pdf.set_x(100)
        pdf.multi_cell(50, 10, f"Text {i} - Page {pdf.page}", 1, "C")
    assert_pdf_equal(pdf, HERE / "break_or_add_page_draw_fill.pdf", tmp_path)


def test_new_page_graphics_state(tmp_path):
    # Make sure that on a page break, all graphics state items are
    # carried over correctly.
    # issue #992 - dash patterns were not handled correctly.
    pdf = fpdf.FPDF()
    pdf.add_page()
    pdf.set_auto_page_break(auto=False)
    pdf.set_font("Courier", style="IU", size=16)
    pdf.set_draw_color(255, 100, 0)
    pdf.set_fill_color(100, 255, 0)
    pdf.set_text_color(0, 0, 255)
    # underline -> set_font()
    # font_style -> set_font()
    pdf.set_stretching(50)
    pdf.set_char_spacing(5)
    # font_family -> set_font()
    # font_size_pt -> set_font()
    # current_font -> set_font()
    pdf.set_dash_pattern(dash=2, gap=4)
    pdf.set_line_width(2)
    # text_mode -> applied via Fragments
    # char lift/scale -> applied via Fragments
    # text_shaping -> applied in the creation of Fragments

    def draw_stuff():
        pdf.cell(
            text="This text is blue, italic, underlined, and squished with wide char spacing.",
            new_x="LEFT",
            new_y="NEXT",
        )
        pdf.cell(text="The box below is green, with a thick, dashed, orange border.")
        pdf.rect(50, 50, 60, 30, style="DF")

    draw_stuff()
    pdf.add_page()
    draw_stuff()
    assert_pdf_equal(pdf, HERE / "new_page_graphics_state.pdf", tmp_path)



"""
This module provides support for embedding and rendering various color font formats
in PDF documents using Type 3 fonts. It defines classes and utilities to handle
different color font technologies, including:

- COLRv0 and COLRv1 (OpenType color vector fonts)
- CBDT/CBLC (bitmap color fonts)
- SBIX (bitmap color fonts)
- SVG (fonts with embedded SVG glyphs)
"""

# muting pyright due to too many fontTools issues
# pyright: reportAttributeAccessIssue=false, reportUnknownVariableType=false, reportPrivateUsage=false
# pyright: reportUnknownMemberType=false, reportUnknownArgumentType=false, reportAssignmentType=false

import logging
import math
from collections import UserList
from io import BytesIO
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    Protocol,
    Sequence,
    Union,
    cast,
)

from fontTools.ttLib.tables.BitmapGlyphMetrics import BigGlyphMetrics, SmallGlyphMetrics
from fontTools.ttLib.tables.C_O_L_R_ import table_C_O_L_R_

# pylint: disable=no-name-in-module
from fontTools.ttLib.tables.otTables import (
    ClipBoxFormat,
    CompositeMode,
    Paint,
    PaintFormat,
    VarAffine2x3,
    VarColorLine,
    VarColorStop,
)
from fontTools.varLib.varStore import VarStoreInstancer

from .drawing import (
    BoundingBox,
    ClippingPath,
    GlyphPathPen,
    GradientPaint,
    GraphicsStyle,
    GraphicsContext,
    ImageSoftMask,
    PaintBlendComposite,
    PaintComposite,
    PaintedPath,
)
from .drawing_primitives import DeviceCMYK, DeviceGray, DeviceRGB, Transform
from .enums import (
    BlendMode,
    CompositingOperation,
    GradientSpreadMethod,
    GradientUnits,
    PathPaintRule,
)
from .pattern import SweepGradient, shape_linear_gradient, shape_radial_gradient

try:
    from PIL import Image
except ImportError:
    Image = None  # type: ignore[assignment]

if TYPE_CHECKING:
    from .fonts import TTFFont
    from .fpdf import FPDF
    from .svg import SVGObject

LOGGER = logging.getLogger(__name__)

PAINT_VAR_MAPPING = {
    PaintFormat.PaintVarSolid: PaintFormat.PaintSolid,
    PaintFormat.PaintVarLinearGradient: PaintFormat.PaintLinearGradient,
    PaintFormat.PaintVarRadialGradient: PaintFormat.PaintRadialGradient,
    PaintFormat.PaintVarSweepGradient: PaintFormat.PaintSweepGradient,
    PaintFormat.PaintVarTransform: PaintFormat.PaintTransform,
    PaintFormat.PaintVarTranslate: PaintFormat.PaintTranslate,
    PaintFormat.PaintVarScale: PaintFormat.PaintScale,
    PaintFormat.PaintVarScaleAroundCenter: PaintFormat.PaintScaleAroundCenter,
    PaintFormat.PaintVarScaleUniform: PaintFormat.PaintScaleUniform,
    PaintFormat.PaintVarScaleUniformAroundCenter: PaintFormat.PaintScaleUniformAroundCenter,
    PaintFormat.PaintVarRotate: PaintFormat.PaintRotate,
    PaintFormat.PaintVarRotateAroundCenter: PaintFormat.PaintRotateAroundCenter,
    PaintFormat.PaintVarSkew: PaintFormat.PaintSkew,
    PaintFormat.PaintVarSkewAroundCenter: PaintFormat.PaintSkewAroundCenter,
}


class Type3FontGlyph:
    # RAM usage optimization:
    __slots__ = (
        "obj_id",
        "glyph_id",
        "unicode",
        "glyph_name",
        "glyph_width",
        "glyph",
        "_glyph_bounds",
    )
    obj_id: int
    glyph_id: int
    unicode: int
    glyph_name: str
    glyph_width: int
    glyph: str
    _glyph_bounds: tuple[int, int, int, int]

    def __init__(self) -> None:
        pass

    def __hash__(self) -> int:
        return self.glyph_id


class Type3Font:

    def __init__(self, fpdf: "FPDF", base_font: "TTFFont"):
        self.i: int = 1
        self.type: str = "type3"
        self.fpdf: "FPDF" = fpdf
        self.base_font: "TTFFont" = base_font
        self.upem: int = self.base_font.ttfont["head"].unitsPerEm
        self.scale: float = 1000 / self.upem  # pyright: ignore[reportUnknownMemberType]
        self.images_used: set[int] = set()
        self.graphics_style_used: set[str] = set()
        self.patterns_used: set[str] = set()
        self.glyphs: list[Type3FontGlyph] = []

    def get_notdef_glyph(self, glyph_id: int) -> Type3FontGlyph:
        notdef = Type3FontGlyph()
        notdef.glyph_id = glyph_id
        notdef.unicode = glyph_id
        notdef.glyph_name = ".notdef"
        notdef.glyph_width = self.base_font.ttfont["hmtx"].metrics[".notdef"][0]
        notdef.glyph = f"{round(notdef.glyph_width * self.scale + 0.001)} 0 d0"
        return notdef

    def get_space_glyph(self, glyph_id: int) -> Type3FontGlyph:
        space = Type3FontGlyph()
        space.glyph_id = glyph_id
        space.unicode = 0x20
        space.glyph_name = "space"
        w = (
            self.base_font.ttfont["hmtx"].metrics["space"][0]
            if "space" in self.base_font.ttfont["hmtx"].metrics
            else self.base_font.ttfont["hmtx"].metrics[".notdef"][0]
        )
        space.glyph_width = round(w + 0.001)
        space.glyph = f"{round(space.glyph_width * self.scale + 0.001)} 0 d0"
        return space

    def load_glyphs(self) -> None:
        WHITES = {
            0x0009,
            0x000A,
            0x000C,
            0x000D,
            0x0020,
            0x00A0,
            0x1680,
            0x2000,
            0x2001,
            0x2002,
            0x2003,
            0x2004,
            0x2005,
            0x2006,
            0x2007,
            0x2008,
            0x2009,
            0x200A,
            0x202F,
            0x205F,
            0x3000,
        }
        for glyph, char_id in self.base_font.subset.items():
            if glyph is None:
                continue
            if glyph.unicode in WHITES or glyph.glyph_name in ("space", "uni00A0"):
                self.glyphs.append(self.get_space_glyph(char_id))
                continue
            if not self.glyph_exists(glyph.glyph_name):
                if self.glyph_exists(".notdef"):
                    self.add_glyph(".notdef", char_id)
                    continue
                self.glyphs.append(self.get_notdef_glyph(char_id))
                continue
            self.add_glyph(glyph.glyph_name, char_id)

    def add_glyph(self, glyph_name: str, char_id: int) -> None:
        g = Type3FontGlyph()
        g.glyph_id = char_id
        g.unicode = char_id
        g.glyph_name = glyph_name
        self.load_glyph_image(g)
        self.glyphs.append(g)

    @classmethod
    def get_target_ppem(cls, font_size_pt: float) -> float:
        # Calculating the target ppem:
        # https://learn.microsoft.com/en-us/typography/opentype/spec/ttch01#display-device-characteristics
        # ppem = point_size * dpi / 72
        # The default PDF dpi resolution is 72 dpi - and we have the 72 dpi hardcoded on our scale factor,
        # so we can simplify the calculation.
        return font_size_pt

    def load_glyph_image(self, glyph: Type3FontGlyph) -> None:
        raise NotImplementedError("Method must be implemented on child class")

    def glyph_exists(self, glyph_name: str) -> bool:
        raise NotImplementedError("Method must be implemented on child class")


class SVGColorFont(Type3Font):
    """Support for SVG OpenType vector color fonts."""

    def glyph_exists(self, glyph_name: str) -> bool:
        glyph_id = self.base_font.ttfont.getGlyphID(glyph_name)
        return any(
            svg_doc.startGlyphID <= glyph_id <= svg_doc.endGlyphID
            for svg_doc in self.base_font.ttfont["SVG "].docList
        )

    def load_glyph_image(self, glyph: Type3FontGlyph) -> None:
        glyph_id = self.base_font.ttfont.getGlyphID(glyph.glyph_name)
        glyph_svg_data = None
        for svg_doc in self.base_font.ttfont["SVG "].docList:
            if svg_doc.startGlyphID <= glyph_id <= svg_doc.endGlyphID:
                glyph_svg_data = svg_doc.data.encode("utf-8")
                break
        if not glyph_svg_data:
            raise ValueError(
                f"Glyph {glyph.glyph_name} (ID: {glyph_id}) not found in SVG font."
            )
        bio = BytesIO(glyph_svg_data)
        bio.seek(0)
        _, img, _ = self.fpdf.preload_glyph_image(glyph_image_bytes=bio)
        if TYPE_CHECKING:
            assert isinstance(img, SVGObject)
        w = round(self.base_font.ttfont["hmtx"].metrics[glyph.glyph_name][0] + 0.001)
        img.base_group.transform = Transform.scaling(self.scale, self.scale)
        output_stream = self.fpdf.draw_vector_glyph(img.base_group, self)
        glyph.glyph = f"{round(w * self.scale)} 0 d0\n" "q\n" f"{output_stream}\n" "Q"
        glyph.glyph_width = w


class ColrV0Layer(Protocol):
    name: str
    colorID: int


class ColrV1Paint(Protocol):
    Paint: Paint


class COLRFont(Type3Font):
    """
    Support for COLRv0 and COLRv1 OpenType color vector fonts.
    https://learn.microsoft.com/en-us/typography/opentype/spec/colr

    COLRv0 is a sequence of glyphs layers with color specification
    and they are built one on top of the other.

    COLRv1 allows for more complex color glyphs by including gradients,
    transformations, and composite operations.

    This class handles both versions of the COLR table by using the
    drawing API to render the glyphs as vector graphics.
    """

    def __init__(
        self, fpdf: "FPDF", base_font: "TTFFont", palette_index: int = 0
    ) -> None:
        super().__init__(fpdf, base_font)
        colr_table: table_C_O_L_R_ = self.base_font.ttfont["COLR"]
        self.colrv0_glyphs: dict[str, tuple[ColrV0Layer]] = {}
        self.colrv1_glyphs: dict[str, ColrV1Paint] = {}
        self.version = colr_table.version
        self.colrv1_clip_boxes = {}
        self.colr_var_instancer = None
        self.colr_var_index_map = None
        if colr_table.version == 0:
            self.colrv0_glyphs = colr_table.ColorLayers
        else:
            try:
                self.colrv0_glyphs = (
                    colr_table._decompileColorLayersV0(colr_table.table) or {}
                )
            except (KeyError, AttributeError, TypeError, ValueError):
                self.colrv0_glyphs = {}
            colr_table_v1 = colr_table.table
            var_store = getattr(colr_table_v1, "VarStore", None)
            if var_store is not None:
                axis_tags = []
                if "fvar" in self.base_font.ttfont:
                    axis_tags = [
                        axis.axisTag for axis in self.base_font.ttfont["fvar"].axes
                    ]
                self.colr_var_instancer = VarStoreInstancer(var_store, axis_tags)
                self.colr_var_instancer.setLocation({tag: 0.0 for tag in axis_tags})
                var_index_map = getattr(colr_table_v1, "VarIndexMap", None)
                if var_index_map is not None:
                    self.colr_var_index_map = var_index_map.mapping
            self.colrv1_glyphs = {
                glyph.BaseGlyph: glyph
                for glyph in colr_table_v1.BaseGlyphList.BaseGlyphPaintRecord
            }
            clip_list = getattr(colr_table_v1, "ClipList", None)
            if clip_list is not None:
                for glyph_name, clip in getattr(clip_list, "clips", {}).items():
                    resolved = self._resolve_clip_box(clip)
                    if resolved is not None:
                        self.colrv1_clip_boxes[glyph_name] = resolved
        self.palette = None
        if "CPAL" in self.base_font.ttfont:
            num_palettes = len(self.base_font.ttfont["CPAL"].palettes)
            # Validate palette index
            if palette_index >= num_palettes:
                LOGGER.warning(
                    "Palette index %s is out of range. This font has %s palettes. Using palette 0.",
                    palette_index,
                    num_palettes,
                )
                palette_index = 0
            palette = self.base_font.ttfont["CPAL"].palettes[palette_index]
            self.palette = [
                (
                    color.red / 255,
                    color.green / 255,
                    color.blue / 255,
                    color.alpha / 255,
                )
                for color in palette
            ]

    def metric_bbox(self) -> BoundingBox:
        return BoundingBox(
            self.base_font.ttfont["head"].xMin,
            self.base_font.ttfont["head"].yMin,
            self.base_font.ttfont["head"].xMax,
            self.base_font.ttfont["head"].yMax,
        )

    def glyph_exists(self, glyph_name: str) -> bool:
        return glyph_name in self.colrv0_glyphs or glyph_name in self.colrv1_glyphs

    def load_glyph_image(self, glyph: Type3FontGlyph) -> None:
        w = round(self.base_font.ttfont["hmtx"].metrics[glyph.glyph_name][0] + 0.001)
        if glyph.glyph_name in self.colrv0_glyphs:
            glyph_layers = self.colrv0_glyphs[glyph.glyph_name]
            img = self.draw_glyph_colrv0(glyph_layers)
        else:
            if self.version < 1 or glyph.glyph_name not in self.colrv1_glyphs:
                raise NotImplementedError(
                    f"No COLRv0 layers and no COLRv1 paint found for '{glyph.glyph_name}'."
                )
            img = self.draw_glyph_colrv1(glyph.glyph_name)
        img.transform = Transform.scaling(self.scale, -self.scale)
        output_stream = self.fpdf.draw_vector_glyph(img, self)
        glyph.glyph = f"{round(w * self.scale)} 0 d0\n" "q\n" f"{output_stream}\n" "Q"
        glyph.glyph_width = w

    def get_color(self, color_index: int, alpha: float = 1) -> DeviceRGB:
        if self.palette is None:  # should never happen
            return DeviceRGB(0, 0, 0, 1)
        if color_index == 0xFFFF:
            # Palette entry 0xFFFF requests the application text foreground color.
            text_color = getattr(self.fpdf, "text_color", DeviceGray(0))
            if isinstance(text_color, DeviceRGB):
                r, g, b = text_color.r, text_color.g, text_color.b
                a = 1.0 if text_color.a is None else text_color.a
            elif isinstance(text_color, DeviceGray):
                r = g = b = text_color.g
                a = 1.0 if text_color.a is None else text_color.a
            elif isinstance(text_color, DeviceCMYK):
                c, m, y, k = text_color.c, text_color.m, text_color.y, text_color.k
                r = 1.0 - min(1.0, c + k)
                g = 1.0 - min(1.0, m + k)
                b = 1.0 - min(1.0, y + k)
                a = 1.0 if text_color.a is None else text_color.a
            else:
                r = g = b = 0.0
                a = 1.0
        else:
            r, g, b, a = self.palette[color_index]
        a *= alpha
        return DeviceRGB(r, g, b, a)

    def draw_glyph_colrv0(self, layers: Sequence[ColrV0Layer]) -> GraphicsContext:
        gc = GraphicsContext()
        for layer in layers:
            path = PaintedPath()
            glyph_set = self.base_font.ttfont.getGlyphSet()
            pen = GlyphPathPen(path, glyphSet=glyph_set)
            glyph = glyph_set[layer.name]
            glyph.draw(pen)
            path.style.fill_color = self.get_color(layer.colorID)
            path.style.stroke_color = self.get_color(layer.colorID)
            gc.add_item(item=path, clone=False)
        return gc

    def draw_glyph_colrv1(self, glyph_name: str) -> GraphicsContext:
        gc = GraphicsContext()
        clip_path = self._build_clip_path(glyph_name)
        if clip_path is not None:
            gc.clipping_path = clip_path
        glyph = self.colrv1_glyphs[glyph_name]
        self.draw_colrv1_paint(
            paint=glyph.Paint,
            parent=gc,
            target_path=None,
            ctm=Transform.identity(),
            visited_glyphs=set(glyph_name),
        )
        return gc

    # pylint: disable=too-many-return-statements
    def draw_colrv1_paint(
        self,
        paint: Paint,
        parent: GraphicsContext,
        target_path: Optional[PaintedPath] = None,
        ctm: Optional[Transform] = None,
        visited_glyphs: Optional[set[str]] = None,
    ) -> tuple[GraphicsContext, Optional[PaintedPath]]:
        """
        Draw a COLRv1 Paint object into the given GraphicsContext.
        This is an implementation of the COLR version 1 rendering algorithm:
        https://learn.microsoft.com/en-us/typography/opentype/spec/colr#colr-version-1-rendering-algorithm
        """
        paint = self._unwrap_paint(paint)
        ctm = ctm or Transform.identity()

        if visited_glyphs is None:
            visited_glyphs = set()

        if paint.Format == PaintFormat.PaintColrLayers:
            layer_list = self.base_font.ttfont["COLR"].table.LayerList
            group = GraphicsContext()
            for layer in range(
                paint.FirstLayerIndex, paint.FirstLayerIndex + paint.NumLayers
            ):
                self.draw_colrv1_paint(
                    paint=layer_list.Paint[layer],
                    parent=group,
                    ctm=ctm,
                    visited_glyphs=visited_glyphs,
                )
            parent.add_item(item=group, clone=False)
            return parent, target_path

        if paint.Format in (
            PaintFormat.PaintSolid,
            PaintFormat.PaintVarSolid,
        ):
            target_path = target_path or self.get_paint_surface()
            target_path.style.fill_color = self.get_color(
                color_index=paint.PaletteIndex, alpha=paint.Alpha
            )
            target_path.style.stroke_color = None
            target_path.style.paint_rule = PathPaintRule.FILL_NONZERO
            return parent, target_path

        if paint.Format == PaintFormat.PaintLinearGradient:
            stops: list[tuple[float, DeviceRGB]] = [
                (stop.StopOffset, self.get_color(stop.PaletteIndex, stop.Alpha))
                for stop in paint.ColorLine.ColorStop
            ]
            if paint.ColorLine.Extend == 2:  # REFLECT
                spread_method = GradientSpreadMethod.REFLECT
            elif paint.ColorLine.Extend == 1:  # REPEAT
                spread_method = GradientSpreadMethod.REPEAT
            else:  # PAD
                spread_method = GradientSpreadMethod.PAD
            linear_gradient = shape_linear_gradient(
                paint.x0,
                paint.y0,
                paint.x1,
                paint.y1,
                stops,
            )
            target_path = target_path or self.get_paint_surface()
            target_path.style.fill_color = GradientPaint(
                gradient=linear_gradient,
                units=GradientUnits.USER_SPACE_ON_USE,
                gradient_transform=ctm,
                apply_page_ctm=False,
                spread_method=spread_method,
            )
            target_path.style.stroke_color = None
            target_path.style.paint_rule = PathPaintRule.FILL_NONZERO
            return parent, target_path

        if paint.Format == PaintFormat.PaintRadialGradient:
            raw = [
                (cs.StopOffset, self.get_color(cs.PaletteIndex, cs.Alpha))
                for cs in paint.ColorLine.ColorStop
            ]
            t_min, t_max, norm_stops = _normalize_color_line(raw)
            c0: tuple[float, float] = (paint.x0, paint.y0)
            r0: float = paint.r0
            c1: tuple[float, float] = (paint.x1, paint.y1)
            r1: float = paint.r1
            fx, fy = _lerp_pt(c0, c1, t_min)
            cx, cy = _lerp_pt(c0, c1, t_max)
            fr = max(_lerp(r0, r1, t_min), 0.0)
            r = max(_lerp(r0, r1, t_max), 1e-6)
            if paint.ColorLine.Extend == 2:  # REFLECT
                spread_method = GradientSpreadMethod.REFLECT
            elif paint.ColorLine.Extend == 1:  # REPEAT
                spread_method = GradientSpreadMethod.REPEAT
            else:  # PAD
                spread_method = GradientSpreadMethod.PAD
            radial_gradient = shape_radial_gradient(
                cx=cx,
                cy=cy,
                r=r,
                fx=fx,
                fy=fy,
                fr=fr,
                stops=norm_stops,
            )
            target_path = target_path or self.get_paint_surface()
            target_path.style.fill_color = GradientPaint(
                gradient=radial_gradient,
                units=GradientUnits.USER_SPACE_ON_USE,
                gradient_transform=ctm,
                apply_page_ctm=False,
                spread_method=spread_method,
            )
            target_path.style.stroke_color = None
            target_path.style.paint_rule = PathPaintRule.FILL_NONZERO
            return parent, target_path

        if paint.Format == PaintFormat.PaintSweepGradient:  # 8
            stops = [
                (cs.StopOffset, self.get_color(cs.PaletteIndex, cs.Alpha))
                for cs in paint.ColorLine.ColorStop
            ]

            if paint.ColorLine.Extend == 2:  # REFLECT
                spread_method = GradientSpreadMethod.REFLECT
            elif paint.ColorLine.Extend == 1:  # REPEAT
                spread_method = GradientSpreadMethod.REPEAT
            else:
                spread_method = GradientSpreadMethod.PAD

            cx = paint.centerX
            cy = paint.centerY

            # COLRv1 defines sweep angles clockwise from the positive X axis.
            # We build gradients in glyph space, which later undergoes a Y-axis flip
            # when emitted to PDF coordinates. To compensate, convert the COLR angles
            # directly to mathematical radians (counter-clockwise); the subsequent flip
            # restores the expected clockwise visual direction.
            start_angle, end_angle = self._sweep_angles(
                paint.startAngle, paint.endAngle
            )

            # Build a lazy sweep gradient object (bbox-resolved at emit time)
            sweep_gradient = SweepGradient(
                cx=cx,
                cy=cy,
                start_angle=start_angle,
                end_angle=end_angle,
                stops=stops,
                spread_method=spread_method,
                segments=None,
                inner_radius_factor=0.002,
            )

            target_path = target_path or self.get_paint_surface()
            target_path.style.fill_color = GradientPaint(
                gradient=sweep_gradient,
                units=GradientUnits.USER_SPACE_ON_USE,
                gradient_transform=ctm,
                apply_page_ctm=False,
                spread_method=spread_method,
            )
            target_path.style.stroke_color = None
            target_path.style.paint_rule = PathPaintRule.FILL_NONZERO
            return parent, target_path

        if paint.Format == PaintFormat.PaintGlyph:  # 10
            glyph_set = self.base_font.ttfont.getGlyphSet()
            clipping_path = ClippingPath()
            glyph_set[paint.Glyph].draw(GlyphPathPen(clipping_path, glyphSet=glyph_set))
            clipping_path.transform = (
                clipping_path.transform or Transform.identity()
            ) @ ctm

            if getattr(paint, "Paint", None) is None:
                return parent, None

            group = GraphicsContext()
            group.clipping_path = clipping_path

            group, surface_path = self.draw_colrv1_paint(
                paint=paint.Paint,
                parent=group,
                ctm=Transform.identity(),
                visited_glyphs=visited_glyphs,
            )
            if surface_path is not None:
                group.add_item(item=surface_path, clone=False)
            parent.add_item(item=group, clone=False)
            return parent, None

        if paint.Format == PaintFormat.PaintColrGlyph:
            ref: str = getattr(paint, "Glyph", None) or getattr(paint, "GlyphID", None)  # type: ignore[assignment]
            if isinstance(ref, int):
                ref_name = self.base_font.ttfont.getGlyphName(ref)
            else:
                ref_name = ref
            if ref_name in visited_glyphs:
                LOGGER.warning("Skipping recursive COLR glyph reference '%s'", ref_name)
                return parent, target_path  # nothing to draw
            rec = self.colrv1_glyphs.get(ref_name)
            if rec is None or getattr(rec, "Paint", None) is None:
                return parent, target_path  # nothing to draw

            visited_glyphs.add(ref_name)
            try:
                group = GraphicsContext()
                clip_path = self._build_clip_path(ref_name)
                if clip_path is not None:
                    group.clipping_path = clip_path
                self.draw_colrv1_paint(
                    paint=rec.Paint,
                    parent=group,
                    ctm=ctm,
                    visited_glyphs=visited_glyphs,
                )
                parent.add_item(item=group, clone=False)
            finally:
                visited_glyphs.remove(ref_name)
            return parent, target_path

        if paint.Format in (
            PaintFormat.PaintTransform,  # 12
            PaintFormat.PaintVarTransform,  # 13
            PaintFormat.PaintTranslate,  # 14
            PaintFormat.PaintVarTranslate,  # 15
            PaintFormat.PaintScale,  # 16
            PaintFormat.PaintVarScale,  # 17
            PaintFormat.PaintScaleAroundCenter,  # 18
            PaintFormat.PaintVarScaleAroundCenter,  # 19
            PaintFormat.PaintScaleUniform,  # 20
            PaintFormat.PaintVarScaleUniform,  # 21
            PaintFormat.PaintScaleUniformAroundCenter,  # 22
            PaintFormat.PaintVarScaleUniformAroundCenter,  # 23
            PaintFormat.PaintRotate,  # 24
            PaintFormat.PaintVarRotate,  # 25
            PaintFormat.PaintRotateAroundCenter,  # 26
            PaintFormat.PaintVarRotateAroundCenter,  # 27
            PaintFormat.PaintSkew,  # 28
            PaintFormat.PaintVarSkew,  # 29
            PaintFormat.PaintSkewAroundCenter,  # 30
            PaintFormat.PaintVarSkewAroundCenter,  # 31
        ):
            transform = self._transform_from_paint(paint)
            new_ctm = ctm @ transform
            return self.draw_colrv1_paint(
                paint=paint.Paint,
                parent=parent,
                target_path=target_path,
                ctm=new_ctm,
                visited_glyphs=visited_glyphs,
            )

        if paint.Format in (
            PaintFormat.PaintVarLinearGradient,  # 5
            PaintFormat.PaintVarRadialGradient,  # 7
            PaintFormat.PaintVarSweepGradient,
        ):  # 9
            raise NotImplementedError("Variable fonts are not yet supported.")

        if paint.Format == PaintFormat.PaintComposite:  # 32
            backdrop_node = GraphicsContext()
            _, backdrop_path = self.draw_colrv1_paint(
                paint=paint.BackdropPaint,
                parent=backdrop_node,
                ctm=ctm,
                visited_glyphs=visited_glyphs,
            )
            if backdrop_path is not None:
                backdrop_node.add_item(item=backdrop_path, clone=False)

            source_node = GraphicsContext()
            _, source_path = self.draw_colrv1_paint(
                paint=paint.SourcePaint,
                parent=source_node,
                ctm=ctm,
                visited_glyphs=visited_glyphs,
            )
            if source_path is not None:
                source_node.add_item(item=source_path, clone=False)

            composite_type, composite_mode = self.get_composite_mode(
                paint.CompositeMode
            )
            if composite_type == "Blend":
                if TYPE_CHECKING:
                    assert isinstance(composite_mode, BlendMode)
                parent.add_item(
                    item=PaintBlendComposite(
                        backdrop=backdrop_node,
                        source=source_node,
                        blend_mode=composite_mode,
                    ),
                    clone=False,
                )
            elif composite_type == "Compositing":
                if TYPE_CHECKING:
                    assert isinstance(composite_mode, CompositeMode)
                composite_node = PaintComposite(
                    backdrop=backdrop_node,
                    source=source_node,
                    operation=composite_mode,  # pyright: ignore[reportArgumentType]
                )
                parent.add_item(item=composite_node, clone=False)
            else:
                raise ValueError("Composite operation not supported - {composite_type}")
            return parent, None

        raise NotImplementedError(f"Unknown PaintFormat: {paint.Format}")

    @classmethod
    def _sweep_angles(cls, start_deg: float, end_deg: float) -> tuple[float, float]:
        start_norm = math.fmod(start_deg, 360.0)
        if start_norm < 0.0:
            start_norm += 360.0
        span_deg = math.fmod(end_deg - start_deg, 360.0)
        if span_deg <= 0.0:
            span_deg += 360.0
        start_rad = math.radians(start_norm)
        end_rad = start_rad + math.radians(span_deg)
        return start_rad, end_rad

    @classmethod
    def _transform_from_paint(cls, paint: Paint) -> Transform:
        paint_format = paint.Format
        if paint_format in (PaintFormat.PaintTransform, PaintFormat.PaintVarTransform):
            transform = paint.Transform
            return Transform(
                transform.xx,
                transform.yx,
                transform.xy,
                transform.yy,
                transform.dx,
                transform.dy,
            )
        if paint_format in (PaintFormat.PaintTranslate, PaintFormat.PaintVarTranslate):
            return Transform.translation(paint.dx, paint.dy)
        if paint_format in (PaintFormat.PaintScale, PaintFormat.PaintVarScale):
            return Transform.scaling(paint.scaleX, paint.scaleY)
        if paint_format in (
            PaintFormat.PaintScaleAroundCenter,
            PaintFormat.PaintVarScaleAroundCenter,
        ):
            return Transform.scaling(paint.scaleX, paint.scaleY).about(
                paint.centerX, paint.centerY
            )
        if paint_format in (
            PaintFormat.PaintScaleUniform,
            PaintFormat.PaintVarScaleUniform,
        ):
            return Transform.scaling(paint.scale, paint.scale)
        if paint_format in (
            PaintFormat.PaintScaleUniformAroundCenter,
            PaintFormat.PaintVarScaleUniformAroundCenter,
        ):
            return Transform.scaling(paint.scale, paint.scale).about(
                paint.centerX, paint.centerY
            )
        if paint_format in (PaintFormat.PaintRotate, PaintFormat.PaintVarRotate):
            return Transform.rotation_d(paint.angle)
        if paint_format in (
            PaintFormat.PaintRotateAroundCenter,
            PaintFormat.PaintVarRotateAroundCenter,
        ):
            return Transform.rotation_d(paint.angle).about(paint.centerX, paint.centerY)
        if paint_format in (PaintFormat.PaintSkew, PaintFormat.PaintVarSkew):
            return Transform.skewing_d(-paint.xSkewAngle, paint.ySkewAngle)
        if paint_format in (
            PaintFormat.PaintSkewAroundCenter,
            PaintFormat.PaintVarSkewAroundCenter,
        ):
            return Transform.skewing_d(-paint.xSkewAngle, paint.ySkewAngle).about(
                paint.centerX, paint.centerY
            )
        raise NotImplementedError(f"Transform not implemented for {format}")

    def get_paint_surface(self) -> PaintedPath:
        """
        Creates a surface representing the whole glyph area for actions that require
        painting an infinite surface and clipping to a geometry path
        """
        paint_surface = PaintedPath()
        surface_bbox = self.metric_bbox()
        paint_surface.rectangle(
            x=surface_bbox.x0,
            y=surface_bbox.y0,
            w=surface_bbox.width,
            h=surface_bbox.height,
        )
        return paint_surface

    @classmethod
    def get_composite_mode(
        cls, composite_mode: CompositeMode
    ) -> (
        tuple[Literal["Compositing"], CompositingOperation]
        | tuple[Literal["Blend"], BlendMode]
    ):
        """Get the FPDF BlendMode for a given CompositeMode."""

        map_compositing_operation = {
            CompositeMode.SRC: CompositingOperation.SOURCE,
            CompositeMode.DEST: CompositingOperation.DESTINATION,
            CompositeMode.CLEAR: CompositingOperation.CLEAR,
            CompositeMode.SRC_OVER: CompositingOperation.SOURCE_OVER,
            CompositeMode.DEST_OVER: CompositingOperation.DESTINATION_OVER,
            CompositeMode.SRC_IN: CompositingOperation.SOURCE_IN,
            CompositeMode.DEST_IN: CompositingOperation.DESTINATION_IN,
            CompositeMode.SRC_OUT: CompositingOperation.SOURCE_OUT,
            CompositeMode.DEST_OUT: CompositingOperation.DESTINATION_OUT,
            CompositeMode.SRC_ATOP: CompositingOperation.SOURCE_ATOP,
            CompositeMode.DEST_ATOP: CompositingOperation.DESTINATION_ATOP,
            CompositeMode.XOR: CompositingOperation.XOR,
        }

        compositing_operation = map_compositing_operation.get(composite_mode, None)
        if compositing_operation is not None:
            return ("Compositing", compositing_operation)

        map_blend_mode = {
            CompositeMode.PLUS: BlendMode.SCREEN,  # approximation
            CompositeMode.SCREEN: BlendMode.SCREEN,
            CompositeMode.OVERLAY: BlendMode.OVERLAY,
            CompositeMode.DARKEN: BlendMode.DARKEN,
            CompositeMode.LIGHTEN: BlendMode.LIGHTEN,
            CompositeMode.COLOR_DODGE: BlendMode.COLOR_DODGE,
            CompositeMode.COLOR_BURN: BlendMode.COLOR_BURN,
            CompositeMode.HARD_LIGHT: BlendMode.HARD_LIGHT,
            CompositeMode.SOFT_LIGHT: BlendMode.SOFT_LIGHT,
            CompositeMode.DIFFERENCE: BlendMode.DIFFERENCE,
            CompositeMode.EXCLUSION: BlendMode.EXCLUSION,
            CompositeMode.MULTIPLY: BlendMode.MULTIPLY,
            CompositeMode.HSL_HUE: BlendMode.HUE,
            CompositeMode.HSL_SATURATION: BlendMode.SATURATION,
            CompositeMode.HSL_COLOR: BlendMode.COLOR,
            CompositeMode.HSL_LUMINOSITY: BlendMode.LUMINOSITY,
        }
        blend_mode = map_blend_mode.get(composite_mode, None)
        if blend_mode is not None:
            return ("Blend", blend_mode)

        raise NotImplementedError(f"Unknown composite mode: {composite_mode}")

    def _unwrap_paint(self, paint: Paint) -> Union[Paint, "VarTableWrapper"]:
        mapped_format = PAINT_VAR_MAPPING.get(paint.Format)
        if mapped_format is None or self.colr_var_instancer is None:
            return paint
        return VarTableWrapper(
            paint,
            self.colr_var_instancer,
            self.colr_var_index_map,
            format_override=mapped_format,
        )

    def _build_clip_path(self, glyph_name: str) -> Optional[ClippingPath]:
        clip_box = self.colrv1_clip_boxes.get(glyph_name)
        if clip_box is None:
            return None
        x_min, y_min, x_max, y_max = clip_box
        clip_path = ClippingPath()
        clip_path.move_to(x_min, y_min)
        clip_path.rectangle(x_min, y_min, x_max - x_min, y_max - y_min)
        return clip_path

    def _resolve_clip_box(
        self, clip: Any
    ) -> Optional[tuple[float, float, float, float]]:
        if clip is None:
            return None
        if (
            getattr(clip, "Format", None) == ClipBoxFormat.Variable
            and self.colr_var_instancer is not None
        ):
            clip = VarTableWrapper(
                clip,
                self.colr_var_instancer,
                self.colr_var_index_map,
            )
        if hasattr(clip, "xMin") and hasattr(clip, "xMax"):
            return (clip.xMin, clip.yMin, clip.xMax, clip.yMax)
        LOGGER.debug("Unsupported COLRv1 clip format for clip box")
        return None


class VarTableWrapper:
    def __init__(
        self,
        wrapped: Any,
        instancer: VarStoreInstancer,
        var_index_map: Any = None,
        format_override: Optional[int] = None,
    ) -> None:
        assert not isinstance(wrapped, VarTableWrapper)
        self._wrapped = wrapped
        self._instancer = instancer
        self._var_index_map = var_index_map
        self._format_override = format_override
        self._var_attrs = {
            attr: idx for idx, attr in enumerate(wrapped.getVariableAttrs())
        }

    def __repr__(self) -> str:
        return f"VarTableWrapper({self._wrapped!r})"

    def _get_var_index_for_attr(self, attr_name: str) -> Any:
        offset = self._var_attrs.get(attr_name)
        if offset is None:
            return None
        base_index = self._wrapped.VarIndexBase
        if base_index == 0xFFFFFFFF:
            return base_index
        var_idx = base_index + offset
        if self._var_index_map is not None:
            try:
                var_idx = self._var_index_map[var_idx]
            except IndexError:
                pass
        return var_idx

    def _get_delta_for_attr(self, attr_name: str, var_idx: Any) -> Any:
        delta = self._instancer[var_idx]
        converter = self._wrapped.getConverterByName(attr_name)
        if hasattr(converter, "fromInt"):
            delta = converter.fromInt(delta)
        return delta

    def __getattr__(self, attr_name: str) -> Any:
        if attr_name == "Format" and self._format_override is not None:
            return self._format_override

        value = getattr(self._wrapped, attr_name)

        var_idx = self._get_var_index_for_attr(attr_name)
        if var_idx is not None:
            if var_idx < 0xFFFFFFFF:
                value += self._get_delta_for_attr(attr_name, var_idx)
        elif isinstance(value, (VarAffine2x3, VarColorLine)):
            value = VarTableWrapper(value, self._instancer, self._var_index_map)
        elif (
            isinstance(value, (list, UserList))
            and value
            and isinstance(value[0], VarColorStop)
        ):
            value = [
                VarTableWrapper(item, self._instancer, self._var_index_map)
                for item in value
            ]

        return value


class CBDTColorFont(Type3Font):
    """Support for CBDT+CBLC bitmap color fonts."""

    # Only looking at the first strike - Need to look all strikes available on the CBLC table first?
    def glyph_exists(self, glyph_name: str) -> bool:
        return glyph_name in self.base_font.ttfont["CBDT"].strikeData[0]

    def load_glyph_image(self, glyph: Type3FontGlyph) -> None:
        ppem = self.base_font.ttfont["CBLC"].strikes[0].bitmapSizeTable.ppemX
        g = self.base_font.ttfont["CBDT"].strikeData[0][glyph.glyph_name]
        glyph_bitmap = g.data[9:]
        metrics = g.metrics
        if isinstance(metrics, SmallGlyphMetrics):
            x_min = round(metrics.BearingX * self.upem / ppem)
            y_min = round((metrics.BearingY - metrics.height) * self.upem / ppem)
            x_max = round(metrics.width * self.upem / ppem)
            y_max = round(metrics.BearingY * self.upem / ppem)
        elif isinstance(metrics, BigGlyphMetrics):
            x_min = round(metrics.horiBearingX * self.upem / ppem)
            y_min = round((metrics.horiBearingY - metrics.height) * self.upem / ppem)
            x_max = round(metrics.width * self.upem / ppem)
            y_max = round(metrics.horiBearingY * self.upem / ppem)
        else:  # fallback scenario: use font bounding box
            x_min = self.base_font.ttfont["head"].xMin
            y_min = self.base_font.ttfont["head"].yMin
            x_max = self.base_font.ttfont["head"].xMax
            y_max = self.base_font.ttfont["head"].yMax

        bio = BytesIO(glyph_bitmap)
        bio.seek(0)
        _, _, info = self.fpdf.preload_glyph_image(glyph_image_bytes=bio)
        w = round(self.base_font.ttfont["hmtx"].metrics[glyph.glyph_name][0] + 0.001)
        glyph.glyph = (
            f"{round(w * self.scale)} 0 d0\n"
            "q\n"
            f"{(x_max - x_min)* self.scale} 0 0 {(-y_min + y_max)*self.scale} {x_min*self.scale} {y_min*self.scale} cm\n"
            f"/I{info['i']} Do\nQ"
        )
        self.images_used.add(info["i"])  # type: ignore[arg-type]
        glyph.glyph_width = w


class EBDTBitmapFont(Type3Font):
    """Support for EBLC+EBDT bitmap fonts."""

    def __init__(self, fpdf: "FPDF", base_font: "TTFFont"):
        super().__init__(fpdf, base_font)
        self._glyph_strike_indexes: dict[str, int] = {}

    def _find_glyph_strike_index(self, glyph_name: str) -> Optional[int]:
        strike_index = self._glyph_strike_indexes.get(glyph_name)
        if strike_index is not None:
            return strike_index

        strikes_data = self.base_font.ttfont["EBDT"].strikeData
        strikes = self.base_font.ttfont["EBLC"].strikes
        strike_indexes = [
            i for i, strike_data in enumerate(strikes_data) if glyph_name in strike_data
        ]
        if not strike_indexes:
            return None

        target_ppem = self.get_target_ppem(self.base_font.biggest_size_pt)
        bigger_or_equal = [
            i for i in strike_indexes if strikes[i].bitmapSizeTable.ppemX >= target_ppem
        ]
        if bigger_or_equal:
            strike_index = min(bigger_or_equal, key=lambda i: self._ppem_x(strikes, i))
        else:
            strike_index = max(strike_indexes, key=lambda i: self._ppem_x(strikes, i))
        self._glyph_strike_indexes[glyph_name] = strike_index
        return strike_index

    @staticmethod
    def _ppem_x(strikes: Sequence[Any], strike_index: int) -> int:
        return int(strikes[strike_index].bitmapSizeTable.ppemX)

    def _get_glyph_metrics(
        self, strike_index: int, glyph_name: str, bitmap_glyph: Any
    ) -> Any:
        metrics = getattr(bitmap_glyph, "metrics", None)
        if metrics is not None:
            return metrics
        for index_sub_table in (
            self.base_font.ttfont["EBLC"].strikes[strike_index].indexSubTables
        ):
            if glyph_name not in index_sub_table.names:
                continue
            metrics = getattr(index_sub_table, "metrics", None)
            if metrics is not None:
                return metrics
            break
        return None

    @classmethod
    def _decode_row(cls, packed_row: bytes, width: int, bit_depth: int) -> bytearray:
        max_value = (1 << bit_depth) - 1
        row_values = bytearray(width)
        bit_index = 0
        for pixel_index in range(width):
            byte_index = bit_index // 8
            bit_offset = bit_index % 8
            bits_in_first_byte = min(bit_depth, 8 - bit_offset)
            if bits_in_first_byte == bit_depth:
                shift = 8 - bit_offset - bit_depth
                value = (packed_row[byte_index] >> shift) & max_value
            else:
                first = packed_row[byte_index] & ((1 << bits_in_first_byte) - 1)
                second_bits = bit_depth - bits_in_first_byte
                second = packed_row[byte_index + 1] >> (8 - second_bits)
                value = (first << second_bits) | second
            row_values[pixel_index] = round(value * 255 / max_value)
            bit_index += bit_depth
        return row_values

    @classmethod
    def _bitmap_to_alpha(
        cls,
        bitmap_glyph: Any,
        metrics: Any,
        bit_depth: int,
    ) -> bytes:
        alpha = bytearray(metrics.width * metrics.height)
        for row_index in range(metrics.height):
            packed_row = bitmap_glyph.getRow(
                row_index, bitDepth=bit_depth, metrics=metrics
            )
            row = cls._decode_row(packed_row, metrics.width, bit_depth)
            start = row_index * metrics.width
            alpha[start : start + metrics.width] = row
        return bytes(alpha)

    def glyph_exists(self, glyph_name: str) -> bool:
        return self._find_glyph_strike_index(glyph_name) is not None

    def load_glyph_image(self, glyph: Type3FontGlyph) -> None:
        if Image is None:
            raise EnvironmentError(
                f"{glyph.glyph_name}: Pillow is required to render EBDT glyphs."
            )

        strike_index = self._find_glyph_strike_index(glyph.glyph_name)
        if strike_index is None:
            raise ValueError(f"{glyph.glyph_name}: glyph not found in EBDT strikes.")

        strike = self.base_font.ttfont["EBLC"].strikes[strike_index]
        bit_depth = strike.bitmapSizeTable.bitDepth
        if bit_depth not in (1, 2, 4, 8):
            raise NotImplementedError(
                f"{glyph.glyph_name}: unsupported EBDT bit depth {bit_depth}."
            )

        bitmap_glyph = self.base_font.ttfont["EBDT"].strikeData[strike_index][
            glyph.glyph_name
        ]
        metrics = self._get_glyph_metrics(strike_index, glyph.glyph_name, bitmap_glyph)
        if metrics is None:
            raise NotImplementedError(
                f"{glyph.glyph_name}: EBDT glyph metrics could not be resolved."
            )
        if not hasattr(bitmap_glyph, "getRow"):
            raise NotImplementedError(
                f"{glyph.glyph_name}: unsupported EBDT glyph format ({type(bitmap_glyph).__name__})."
            )

        ppem_x = strike.bitmapSizeTable.ppemX or 1
        ppem_y = strike.bitmapSizeTable.ppemY or ppem_x
        if isinstance(metrics, SmallGlyphMetrics):
            x_min = round(metrics.BearingX * self.upem / ppem_x)
            y_min = round((metrics.BearingY - metrics.height) * self.upem / ppem_y)
            x_max = round((metrics.BearingX + metrics.width) * self.upem / ppem_x)
            y_max = round(metrics.BearingY * self.upem / ppem_y)
        elif isinstance(metrics, BigGlyphMetrics):
            x_min = round(metrics.horiBearingX * self.upem / ppem_x)
            y_min = round((metrics.horiBearingY - metrics.height) * self.upem / ppem_y)
            x_max = round((metrics.horiBearingX + metrics.width) * self.upem / ppem_x)
            y_max = round(metrics.horiBearingY * self.upem / ppem_y)
        else:  # fallback scenario: use font bounding box
            x_min = self.base_font.ttfont["head"].xMin
            y_min = self.base_font.ttfont["head"].yMin
            x_max = self.base_font.ttfont["head"].xMax
            y_max = self.base_font.ttfont["head"].yMax

        w = round(self.base_font.ttfont["hmtx"].metrics[glyph.glyph_name][0] + 0.001)
        if bit_depth == 1:
            alpha = self._bitmap_to_alpha(bitmap_glyph, metrics, bit_depth)
            pixel_w = (x_max - x_min) / max(metrics.width, 1)
            pixel_h = (y_max - y_min) / max(metrics.height, 1)
            path_cmds: list[str] = []
            for row_index in range(metrics.height):
                row_start = row_index * metrics.width
                row = alpha[row_start : row_start + metrics.width]
                col = 0
                while col < metrics.width:
                    if row[col] == 0:
                        col += 1
                        continue
                    start = col
                    while col < metrics.width and row[col] != 0:
                        col += 1
                    run_len = col - start
                    x = (x_min + start * pixel_w) * self.scale
                    y = (
                        y_min + (metrics.height - row_index - 1) * pixel_h
                    ) * self.scale
                    w_run = (run_len * pixel_w) * self.scale
                    h_run = pixel_h * self.scale
                    path_cmds.append(f"{x:.3f} {y:.3f} {w_run:.3f} {h_run:.3f} re")
            if path_cmds:
                glyph.glyph = (
                    f"{round(w * self.scale)} 0 d0\n"
                    "q\n"
                    f"{' '.join(path_cmds)} f\n"
                    "Q"
                )
            else:
                glyph.glyph = f"{round(w * self.scale)} 0 d0"
            glyph.glyph_width = w
            return

        alpha = self._bitmap_to_alpha(bitmap_glyph, metrics, bit_depth)
        alpha_image = Image.frombytes("L", (metrics.width, metrics.height), alpha)
        bio = BytesIO()
        alpha_image.save(bio, format="PNG")
        bio.seek(0)
        _, _, info = self.fpdf.preload_glyph_image(glyph_image_bytes=bio)

        mask_matrix = Transform(
            a=(x_max - x_min) * self.scale,
            b=0,
            c=0,
            d=(y_max - y_min) * self.scale,
            e=x_min * self.scale,
            f=y_min * self.scale,
        )
        bbox = (
            x_min * self.scale,
            y_min * self.scale,
            x_max * self.scale,
            y_max * self.scale,
        )
        soft_mask = ImageSoftMask(cast(int, info["i"]), bbox, mask_matrix)

        soft_mask.object_id = self.fpdf._resource_catalog.register_soft_mask(  # pylint: disable=protected-access
            soft_mask
        )
        style = GraphicsStyle()
        style.soft_mask = soft_mask
        gs_name = self.fpdf._resource_catalog.register_graphics_style(  # pylint: disable=protected-access
            style
        )
        if gs_name is None:
            raise RuntimeError("Failed to register soft mask graphics state.")
        self.graphics_style_used.add(str(gs_name))

        glyph.glyph = (
            f"{round(w * self.scale)} 0 d0\n"
            "q\n"
            f"/{gs_name} gs\n"
            f"{x_min * self.scale} {y_min * self.scale} "
            f"{(x_max - x_min) * self.scale} {(y_max - y_min) * self.scale} re f\n"
            "Q"
        )
        glyph.glyph_width = w


class SBIXColorFont(Type3Font):
    """Support for SBIX bitmap color fonts."""

    def glyph_exists(self, glyph_name: str) -> bool:
        glyph = (
            self.base_font.ttfont["sbix"]
            .strikes[self.get_strike_index()]
            .glyphs.get(glyph_name)
        )
        return glyph is not None and glyph.graphicType is not None

    def get_strike_index(self) -> int:
        target_ppem = self.get_target_ppem(self.base_font.biggest_size_pt)
        ppem_list: list[int] = [
            ppem
            for ppem in self.base_font.ttfont["sbix"].strikes.keys()
            if ppem >= target_ppem
        ]
        if not ppem_list:
            return max(list(self.base_font.ttfont["sbix"].strikes.keys()))  # type: ignore[no-any-return]
        return min(ppem_list)

    def load_glyph_image(self, glyph: Type3FontGlyph) -> None:
        ppem = self.get_strike_index()
        sbix_glyph = (
            self.base_font.ttfont["sbix"].strikes[ppem].glyphs.get(glyph.glyph_name)
        )
        if sbix_glyph.graphicType == "dupe":
            raise NotImplementedError(
                f"{glyph.glyph_name}: Dupe SBIX graphic type not implemented."
            )
            # waiting for an example to test
            # dupe_char = font.getBestCmap()[glyph.imageData]
            # return self.get_color_glyph(dupe_char)

        if sbix_glyph.graphicType not in ("jpg ", "png ", "tiff"):  # pdf or mask
            raise NotImplementedError(
                f" {glyph.glyph_name}: Invalid SBIX graphic type {sbix_glyph.graphicType}."
            )

        bio = BytesIO(sbix_glyph.imageData)
        bio.seek(0)
        _, _, info = self.fpdf.preload_glyph_image(glyph_image_bytes=bio)
        w = round(self.base_font.ttfont["hmtx"].metrics[glyph.glyph_name][0] + 0.001)
        glyf_metrics = self.base_font.ttfont["glyf"].get(glyph.glyph_name)
        assert glyf_metrics is not None
        x_min = glyf_metrics.xMin + sbix_glyph.originOffsetX
        x_max = glyf_metrics.xMax + sbix_glyph.originOffsetX
        y_min = glyf_metrics.yMin + sbix_glyph.originOffsetY
        y_max = glyf_metrics.yMax + sbix_glyph.originOffsetY

        glyph.glyph = (
            f"{round(w * self.scale)} 0 d0\n"
            "q\n"
            f"{(x_max - x_min) * self.scale} 0 0 {(-y_min + y_max) * self.scale} {x_min * self.scale} {y_min * self.scale} cm\n"
            f"/I{info['i']} Do\nQ"
        )
        self.images_used.add(info["i"])  # type: ignore[arg-type]
        glyph.glyph_width = w


# pylint: disable=too-many-return-statements
def get_color_font_object(
    fpdf: "FPDF", base_font: "TTFFont", palette_index: int = 0
) -> Union[Type3Font, None]:
    def has_outline_glyphs() -> bool:
        if base_font.is_cff:
            return True
        if "glyf" not in base_font.ttfont:
            return False
        glyph_names = set(base_font.cmap.values())
        if not glyph_names:
            return False
        glyf_table = base_font.ttfont["glyf"]
        return any(
            glyph_name != ".notdef" and glyph_name in glyf_table
            for glyph_name in glyph_names
        )

    if "CBDT" in base_font.ttfont:
        LOGGER.debug("Font %s is a CBLC+CBDT color font", base_font.name)
        return CBDTColorFont(fpdf, base_font)
    if "EBDT" in base_font.ttfont:
        if has_outline_glyphs():
            # Prefer outlines when a font ships both outlines and bitmap strikes.
            LOGGER.debug(
                "Font %s has EBLC+EBDT tables and outline glyphs; preferring outlines",
                base_font.name,
            )
            return None
        LOGGER.debug("Font %s is a EBLC+EBDT color font", base_font.name)
        return EBDTBitmapFont(fpdf, base_font)
    if "COLR" in base_font.ttfont:
        if base_font.ttfont["COLR"].version == 0:
            LOGGER.debug("Font %s is a COLRv0 color font", base_font.name)
        else:
            LOGGER.debug("Font %s is a COLRv1 color font", base_font.name)
        return COLRFont(fpdf, base_font, palette_index)
    if "SVG " in base_font.ttfont:
        LOGGER.debug("Font %s is a SVG color font", base_font.name)
        return SVGColorFont(fpdf, base_font)
    if "sbix" in base_font.ttfont:
        LOGGER.debug("Font %s is a SBIX color font", base_font.name)
        return SBIXColorFont(fpdf, base_font)
    return None


def _lerp(a: float, b: float, t: float) -> float:
    """ "Scalar linear interpolation"""
    return a + (b - a) * t


def _lerp_pt(
    p0: tuple[float, float], p1: tuple[float, float], t: float
) -> tuple[float, float]:
    """2d vector interpolation"""
    return (_lerp(p0[0], p1[0], t), _lerp(p0[1], p1[1], t))


def _normalize_color_line(
    stops: list[tuple[float, DeviceRGB]],
) -> tuple[float, float, list[tuple[float, DeviceRGB]]]:
    # stops: list[(offset, DeviceRGB)]
    s = sorted(((max(0.0, min(1.0, t)), c) for t, c in stops), key=lambda x: x[0])
    # collapse identical offsets (last wins per spec-ish behavior)
    out: list[tuple[float, DeviceRGB]] = []
    for t, c in s:
        if out and abs(out[-1][0] - t) < 1e-6:
            out[-1] = (t, c)
        else:
            out.append((t, c))
    t_min, t_max = out[0][0], out[-1][0]
    if t_max - t_min < 1e-6:
        # degenerate: treat as solid
        return t_min, t_max, [(0.0, out[-1][1])]
    scale = 1.0 / (t_max - t_min)
    renorm = [((t - t_min) * scale, c) for (t, c) in out]
    return t_min, t_max, renorm



#!/usr/bin/env python3
# Script Dependencies:
#    fpdf2
#    livereload
#    xreload
import asyncio, logging, sys
from traceback import print_exc

from fpdf import FPDF
from livereload.watcher import get_watcher_class
from xreload import xreload

OUT_FILEPATH = "fpdf2-demo.pdf"


def build_pdf():
    pdf = FPDF()
    pdf.set_font("Helvetica", size=16)
    pdf.add_page()
    pdf.y += 50
    pdf.multi_cell(
        h=10,
        w=0,
        align="C",
        text="""Hello fpdf2 user!
Launch this script with --watch
and then try to modify this text while the script is running""",
    )
    pdf.output(OUT_FILEPATH)
    print(f"{OUT_FILEPATH} has been rebuilt")


async def start_watch_and_rebuild():
    logging.basicConfig(
        format="%(asctime)s %(name)s [%(levelname)s] %(message)s",
        datefmt="%H:%M:%S",
        level=logging.INFO,
    )
    logging.getLogger("livereload").setLevel(logging.INFO)
    watcher = get_watcher_class()()
    watcher.watch(__file__, build_pdf)
    print("Watcher started...")
    await watch_periodically(watcher)


async def watch_periodically(watcher, delay_secs=0.8):
    try:
        watcher.examine()
    except Exception:
        print_exc()
    await asyncio.sleep(delay_secs)
    xreload(sys.modules[__name__], new_annotations={"XRELOADED": True})
    await asyncio.create_task(watch_periodically(watcher))


# This conditional ensure that the code below
# does not get executed when calling xreload on this module:
if not __annotations__.get("XRELOADED"):
    build_pdf()
    # The --watch mode is very handy when using a PDF reader
    # that performs hot-reloading, like Sumatra PDF Reader:
    if "--watch" in sys.argv:
        asyncio.run(start_watch_and_rebuild())
